import math
import os
from argparse import Namespace
from collections import defaultdict
from torch import nn
import torch
import copy
from ..logger import beam_logger as logger
import numpy as np
from ..nn import BeamOptimizer, BeamScheduler, MultipleScheduler, BeamNN, BeamDDP
from ..utils import (to_device, check_type, recursive_concatenate,
                     beam_device, filter_dict, lazy_property,
                     is_notebook, DataBatch, dictionary_iterator, recursive_clone, set_item_with_tuple_key,
                     check_nvlink)
from ..dataset import UniversalBatchSampler, UniversalDataset, TransformedDataset
from ..experiment import Experiment, BeamReport
from ..path import beam_path, local_copy
from ..core import Processor, MetaBeamInit
from ..logger import beam_kpi, BeamResult
from timeit import default_timer as timer


class NeuralAlgorithm(Processor):

    def __init__(self, hparams, networks=None, optimizers=None, schedulers=None, processors=None, dataset=None,
                 name=None, experiment=None, **kwargs):

        super().__init__(hparams, name=name, **kwargs)

        self.clear_experiment_properties()
        self._experiment = None
        self._device = None
        if experiment is not None:
            self.experiment = experiment

        self.trial = None

        # the following are set by the experiment
        # some experiment hyperparameters

        self.epoch = 0
        self.t0 = timer()

        self.scalers = {}
        self.networks = {}
        self.processors = {}
        self.swa_networks = {}

        self.optimizers = {}
        self.schedulers = {}
        self.swa_schedulers = {}
        self.schedulers_initial_state = {}

        self.optimizers_name_by_id = {}
        self.schedulers_name_by_id = {}
        self.schedulers_flat = {}
        self.optimizers_flat = {}
        self.optimizers_steps = {}
        self.datasets = {}
        self.persistent_dataloaders = defaultdict(dict)
        self.dataloaders = defaultdict(dict)

        self.epoch_length = None
        self.eval_subset = None
        self.objective = None
        self.best_objective = None
        self.best_epoch = None
        self.best_state = False

        self.add_components(networks=networks, optimizers=optimizers, schedulers=schedulers, processors=processors)

        if self.get_hparam('reload_path') is not None:
            self.load_checkpoint(hparams.reload_path, hparams=False)

        if hparams.store_initial_weights:
            self.initial_weights = self.save_checkpoint()

        if dataset is not None:
            self.load_dataset(dataset)

        self.cb_model = None
        self.reporter = None
        self.training = False

    @property
    def elapsed_time(self):
        return timer() - self.t0

    @staticmethod
    def no_experiment_message(property):
        logger.warning(f"{property} is not supported without an active experiment. Set self.experiment = experiment")

    @lazy_property
    def distributed_training(self):
        if self.experiment is None:
            self.no_experiment_message('distributed_training')
            return False
        return self.experiment.distributed_training

    @lazy_property
    def distributed_training_framework(self):
        if self.experiment is None:
            self.no_experiment_message('distributed_training_framework')
            return None
        return self.experiment.distributed_training_framework

    @lazy_property
    def hpo(self):
        if self.experiment is None:
            self.no_experiment_message('hpo')
            return False
        return self.experiment.hpo

    @lazy_property
    def rank(self):
        if self.experiment is None:
            self.no_experiment_message('rank')
            return 0
        return self.experiment.rank

    @lazy_property
    def world_size(self):
        if self.experiment is None:
            self.no_experiment_message('world_size')
            return 1
        return self.experiment.world_size

    @lazy_property
    def enable_tqdm(self):
        return self.get_hparam('enable_tqdm') if (self.get_hparam('tqdm_threshold') == 0
                                                              or not self.get_hparam('enable_tqdm')) else None

    @lazy_property
    def n_epochs(self):
        return self.get_hparam('n_epochs')

    @lazy_property
    def batch_size_train(self):
        return self.get_hparam('batch_size_train')

    @lazy_property
    def batch_size_eval(self):
        return self.get_hparam('batch_size_eval')

    @lazy_property
    def pin_memory(self):
        return self.is_cuda

    @lazy_property
    def half(self):

        # maybe we should add bfloat16 + accelerate
        if self.native_training and self.model_dtype in [torch.float16, torch.bfloat16, torch.complex32]:
            return True

        return False

    @lazy_property
    def autocast_device(self):
        return 'cuda' if self.is_cuda else 'cpu'

    @lazy_property
    def brain(self):
        if self.model_dtype in [torch.bfloat16]:
            return True
        return False

    @lazy_property
    def model_dtype(self):

        if self.amp:
            return None

        model_dtype = self.get_hparam('model_dtype', default='float32')
        if not self.is_cuda and model_dtype == 'float16':
            logger.warning(f'Autocast on CPU is only supported for bfloat16, defaulting to bfloat16.')
            model_dtype = 'bfloat16'

        return self.dtype_mapping(model_dtype)

    @staticmethod
    def dtype_mapping(model_dtype):
        model_mapping = {'float16': torch.float16, 'bfloat16': torch.bfloat16,
                         'float32': torch.float32, 'complex32': torch.complex32, 'complex64': torch.complex64, }
        return model_mapping[model_dtype]

    @staticmethod
    def accelerate_dtype_mapper(model_dtype):
        model_mapping = {torch.float16: 'fp16', torch.bfloat16: 'bf16', torch.float32: 'no'}
        return model_mapping[model_dtype]

    @lazy_property
    def training_framework(self):
        return self.get_hparam('training_framework', default='torch')

    @lazy_property
    def native_training(self):
        return self.training_framework == 'torch'

    @lazy_property
    def mixed_precision_dtype(self):

        if self.native_training:
            return None

        model_dtype = self.get_hparam('model_dtype', default='float32')

        if self.amp:
            if not self.is_cuda and model_dtype == 'float16':
                logger.warning(f'Autocast on CPU is only supported for bfloat16, defaulting to bfloat16.')
                model_dtype = 'bfloat16'

        return self.dtype_mapping(model_dtype)

    @lazy_property
    def amp(self):
        return self.training_framework == 'amp' if self.is_cuda else False

    # @lazy_property
    # def deepspeed(self):
    #     return self.training_framework == 'deepspeed' and self.get_hparam('n_gpus') > 1

    @lazy_property
    def deepspeed(self):
        return self.training_framework == 'deepspeed'

    @lazy_property
    def scaler(self):
        return torch.cuda.amp.GradScaler() if self.amp else None

    @lazy_property
    def swa_epochs(self):
        swa_epochs = 0
        if self.get_hparam('swa') is not None:
            if int(self.get_hparam('swa')) == self.get_hparam('swa'):
                swa_epochs = int(self.get_hparam('swa'))
            else:
                swa_epochs = int(np.round(self.get_hparam('swa') * self.n_epochs))
        return swa_epochs

    def clear_experiment_properties(self):

        self.clear_cache('device', 'distributed_training', 'distributed_training_framework', 'hpo', 'rank', 'world_size', 'enable_tqdm', 'n_epochs',
                            'batch_size_train', 'batch_size_eval', 'pin_memory', 'autocast_device', 'model_dtype', 'amp',
                            'scaler', 'swa_epochs')

    @lazy_property
    def device(self):
        if self.in_cache('accelerator') and self.accelerator.device_placement:
            device = self.accelerator.device
        elif self._device is not None:
            device = self._device
        elif self.experiment is not None and hasattr(self.experiment, 'device'):
            device = beam_device(self.experiment.device)
        else:
            device = beam_device(self.get_hparam('device'))

        return device

    @lazy_property
    def deepspeed_plugin(self):
        deepspeed_plugin = None
        if self.get_hparam('n_gpus') > 1 and self.accelerate:
            from accelerate.utils import DeepSpeedPlugin
            deepspeed_plugin = DeepSpeedPlugin(self.deepspeed_config(target='accelerate'),
                                               gradient_accumulation_steps=self.get_hparam('accumulate'),)

        return deepspeed_plugin

    @lazy_property
    def fsdp_plugin(self):
        return None

    @lazy_property
    def megatron_lm_plugin(self):
        return None

    @lazy_property
    def accelerate_kwargs_handlers(self):
        return None

    @lazy_property
    def gradient_accumulation_plugin(self):
        return None

    @lazy_property
    def accelerate(self):
        return self.training_framework == 'accelerate'

    @lazy_property
    def accelerator(self):

        acc = None
        if self.accelerate:
            from accelerate import Accelerator

            deepspeed_plugin = self.deepspeed_plugin
            device_placement = None if self.deepspeed else self.get_hparam('device_placement')

            if not check_nvlink():
                os.environ['NCCL_P2P_DISABLE'] = '1'
                os.environ['NCCL_IB_DISABLE'] = '1'

            acc = Accelerator(device_placement=device_placement,
                               split_batches=self.get_hparam('split_batches'),
                               mixed_precision=self.accelerate_dtype_mapper(self.mixed_precision_dtype),
                               gradient_accumulation_steps=self.get_hparam('accumulate'),
                               deepspeed_plugin=deepspeed_plugin,
                               fsdp_plugin=self.fsdp_plugin,
                               megatron_lm_plugin=self.megatron_lm_plugin,
                               kwargs_handlers=self.accelerate_kwargs_handlers,
                               cpu='cpu' == self.get_hparam('device'),
                               project_dir=self.experiment.experiment_dir if self.experiment is not None else None,
                               dispatch_batches=self.get_hparam('dispatch_batches'),
                               even_batches=True,
                               step_scheduler_with_optimizer=self.get_hparam('schedulers_steps') == 'iteration',
                               dynamo_backend=self.get_hparam('dynamo_backend'),
                               gradient_accumulation_plugin=self.gradient_accumulation_plugin,
                               )
            self.clear_cache('device')
        return acc

    @property
    def exclude_pickle_attributes(self):
        return ['networks', 'optimizers', 'schedulers', 'processors', 'datasets', 'scaler',
                'swa_networks', 'swa_schedulers', 'schedulers_initial_state', 'optimizers_name_by_id',
                'schedulers_name_by_id', 'schedulers_flat', 'optimizers_flat', 'optimizers_steps']

    @lazy_property
    def train_reporter(self):
        return BeamReport(objective=self.get_hparam('objective'), objective_mode=self.optimization_mode,
                          aux_objectives=['loss'], aux_objectives_modes=['min'])

    def __getattr__(self, item):
        assert item != 'networks', 'Networks are not initialized yet'
        if item in self.networks:
            return self.networks[item]
        else:
            raise AttributeError(f"Algorithm has no attribute {item}")

    @property
    def dataset(self):
        if not len(self.datasets):
            return None
        if len(self.datasets) == 1:
            return list(self.datasets.values())[0]
        return self.datasets

    def set_reporter(self, reporter=None):
        self.reporter = reporter
        self.reporter.reset_time(None)
        self.reporter.reset_epoch(0, total_epochs=None)

    def set_train_reporter(self, first_epoch, n_epochs=None):

        if n_epochs is None:
            n_epochs = self.n_epochs

        self.reporter = self.train_reporter
        self.reporter.reset_time(first_epoch=first_epoch, n_epochs=n_epochs)

    @lazy_property
    def is_notebook(self):
        return is_notebook()

    def to(self, device):

        logger.warning("Current implementation transforms only the networks dictionary. Don't use for training.")
        device = torch.device(device)

        if not self.accelerate or not self.accelerator.device_placement:
            self._device = beam_device(device)
            self.clear_cache('device')

        for net in self.networks.values():
            net.to(device)

        return self

    @classmethod
    def from_pretrained(cls, path, override_hparams=None, hparams=None, dataset=None, alg_args=None, alg_kwargs=None,
                        dataset_args=None, dataset_kwargs=None, reload_iloc=None, reload_loc=None, reload_name=None,
                        device=None, **kwargs):

        if hparams is not None:
            override_hparams = hparams

        if device is not None:
            override_hparams = override_hparams or {}
            override_hparams['device'] = device

        experiment = Experiment.reload_from_path(path, override_hparams=override_hparams, reload_iloc=reload_iloc,
                                                 reload_loc=reload_loc, reload_name=reload_name, **kwargs)

        if alg_args is None and alg_kwargs is None:
            alg_args, alg_kwargs = experiment.get_alg_init_args()

            alg_kwargs.pop('experiment', None)
            dataset_e = alg_kwargs.pop('dataset', None)
            dataset = dataset or dataset_e

            # remove the hparams from the args list
            alg_args = alg_args[1:]

        alg = experiment.algorithm_generator(cls, dataset=dataset, alg_args=alg_args, alg_kwargs=alg_kwargs,
                                             dataset_args=dataset_args, dataset_kwargs=dataset_kwargs,
                                             store_init_args=False)

        if hparams is not None:
            experiment = Experiment(hparams)

        alg.experiment = experiment
        return alg

    def add_networks(self, networks):
        self.add_components(networks=networks)

    def add_processors(self, processors):
        self.add_components(processors=processors)

    def add_optimizers(self, optimizers):
        self.add_components(optimizers=optimizers)

    def add_schedulers(self, schedulers):
        self.add_components(schedulers=schedulers)

    def add_network(self, network, name):
        self.add_components(networks={name: network})

    def add_processor(self, processor, name):
        self.add_components(processors={name: processor})

    def add_optimizer(self, optimizer, name):
        self.add_components(optimizers={name: optimizer})

    def add_scheduler(self, scheduler, name):
        self.add_components(schedulers={name: scheduler})

    def report_scalar(self, name, val, subset=None, aggregation=None, append=None, **kwargs):
        self.reporter.report_scalar(name, val, subset=subset, aggregation=aggregation, append=append, **kwargs)

    def report_data(self, name, val, subset=None, data_type=None, **kwargs):

        if '/' in name:
            dt, name = name.split('/')

            if data_type is None:
                data_type = dt
            else:
                data_type = f"{dt}_{data_type}"

        self.reporter.report_data(name, val, subset=subset, data_type=data_type, **kwargs)

    def report_image(self, name, val, subset=None, **kwargs):
        self.reporter.report_image(name, val, subset=subset, **kwargs)

    def report_images(self, name, val, subset=None, **kwargs):
        self.reporter.report_images(name, val, subset=subset, **kwargs)

    def report_scalars(self, name, val, subset=None, **kwargs):
        self.reporter.report_scalars(name, val, subset=subset, **kwargs)

    def report_histogram(self, name, val, subset=None, **kwargs):
        self.reporter.report_histogram(name, val, subset=subset, **kwargs)

    def report_figure(self, name, val, subset=None, **kwargs):
        self.reporter.report_figure(name, val, subset=subset, **kwargs)

    def report_video(self, name, val, subset=None, **kwargs):
        self.reporter.report_video(name, val, subset=subset, **kwargs)

    def report_audio(self, name, val, subset=None, **kwargs):
        self.reporter.report_audio(name, val, subset=subset, **kwargs)

    def report_embedding(self, name, val, subset=None, **kwargs):
        self.reporter.report_embedding(name, val, subset=subset, **kwargs)

    def report_text(self, name, val, subset=None, **kwargs):
        self.reporter.report_text(name, val, subset=subset, **kwargs)

    def report_mesh(self, name, val, subset=None, **kwargs):
        self.reporter.report_mesh(name, val, subset=subset, **kwargs)

    def report_pr_curve(self, name, val, subset=None, **kwargs):
        self.reporter.report_pr_curve(name, val, subset=subset, **kwargs)

    def get_scalar(self, name, subset=None, aggregate=False):
        v = self.reporter.get_scalar(name, subset=subset, aggregate=aggregate)
        return self.reporter.stack_scalar(v)

    def get_scalars(self, name, subset=None, aggregate=False):
        d = self.reporter.get_scalars(name, subset=subset, aggregate=aggregate)
        for k, v in d.items():
            d[k] = self.reporter.stack_scalar(v)
        return d

    def get_data(self, name, subset=None, data_type=None):

        if '/' in name:
            dt, name = name.split('/')

            if data_type is None:
                data_type = dt
            else:
                data_type = f"{dt}_{data_type}"

        return self.reporter.get_data(name, subset=subset, data_type=data_type)

    def get_image(self, name, subset=None):
        return self.reporter.get_image(name, subset=subset)

    def get_images(self, name, subset=None):
        return self.reporter.get_images(name, subset=subset)

    def get_histogram(self, name, subset=None):
        return self.reporter.get_histogram(name, subset=subset)

    def get_figure(self, name, subset=None):
        return self.reporter.get_figure(name, subset=subset)

    def get_video(self, name, subset=None):
        return self.reporter.get_video(name, subset=subset)

    def get_audio(self, name, subset=None):
        return self.reporter.get_audio(name, subset=subset)

    def get_embedding(self, name, subset=None):
        return self.reporter.get_embedding(name, subset=subset)

    def get_text(self, name, subset=None):
        return self.reporter.get_text(name, subset=subset)

    def get_mesh(self, name, subset=None):
        return self.reporter.get_mesh(name, subset=subset)

    def get_pr_curve(self, name, subset=None):
        return self.reporter.get_pr_curve(name, subset=subset)

    def add_components(self, networks=None, optimizers=None, schedulers=None, processors=None,
                       build_optimizers=True, build_schedulers=True, name='net'):

        if networks is None:
            networks = self.networks
        else:
            if isinstance(networks, nn.Module):
                networks = {name: networks}
            elif check_type(networks).minor == 'dict':
                pass
            else:
                raise NotImplementedError("Network type is unsupported")

            for k, net in networks.items():
                if k in self.networks:
                    self.networks.pop(k)
                    logger.warning(f"Found network with identical keys: {k}. Overriding previous network.")
                    if k in self.optimizers:
                        self.optimizers.pop(k)

                net = self.register_network(networks[k], name=k)

                self.networks[k] = net

                if self.get_hparam('swa') is not None:
                    self.swa_networks[k] = torch.optim.swa_utils.AveragedModel(net)

        if optimizers is None:
            optimizers = {}

        elif isinstance(optimizers, dict):
            for k, o in optimizers.items():
                if callable(o):
                    try:
                        optimizers[k] = o(networks[k])
                    except TypeError:
                        optimizers[k] = o(networks[k].parameters())
                else:
                    o.load_state_dict(o.state_dict())
                    optimizers[k] = o

        elif isinstance(optimizers, torch.optim.Optimizer) or isinstance(optimizers, BeamOptimizer):
            optimizers.load_state_dict(optimizers.state_dict())
            optimizers = {name: optimizers}

        elif callable(optimizers):
            try:
                optimizers = {name: optimizers(networks[name])}
            except TypeError:
                optimizers = {name: optimizers(networks[name].parameters())}
        else:
            raise NotImplementedError

        if build_optimizers:

            momentum = self.get_hparam('momentum')
            if momentum is None:
                momentum = self.get_hparam('beta1')

            for k, v in networks.items():
                if k not in optimizers and not self.deepspeed:
                    optimizers[k] = BeamOptimizer(v, dense_args={'lr': self.get_hparam('lr_dense', specific=k),
                                                                  'weight_decay': self.get_hparam('weight_decay', specific=k),
                                                                  'betas': (self.get_hparam('momentum', specific=k, default=momentum),
                                                                            self.get_hparam('beta2', specific=k)),
                                                                  'eps': self.get_hparam('eps', specific=k),},
                                                   sparse_args={'lr': self.get_hparam('lr_sparse', specific=k),
                                                                'betas': (self.get_hparam('momentum', specific=k, default=momentum),
                                                                          self.get_hparam('beta2', specific=k)),
                                                                'eps': self.get_hparam('eps', specific=k)},
                                                   clip=self.get_hparam('clip_gradient', specific=k), amp=self.amp,
                                                   accumulate=self.get_hparam('accumulate', specific=k),
                                                   model_dtype=self.mixed_precision_dtype)

        if processors is None:
            processors = {}
        elif isinstance(processors, Processor):
            processors = {processors.name: processors}
        elif isinstance(processors, list):
            processors = {p.name: p for p in processors}

        for k, v in processors.items():
            if k in self.processors:
                self.processors.pop(k)
                logger.warning(f"Found processor with identical keys: {k}. Overriding previous processor.")
            self.processors[k] = v

        if schedulers is None:
            schedulers = {}

        for k, opt in optimizers.items():

            self.optimizers[k] = opt

            if self.get_hparam('swa') is not None:

                kwargs = {'anneal_epochs': self.get_hparam('swa_anneal_epochs', specific=k), 'anneal_strategy': 'cos'}

                if type(opt) is BeamOptimizer:
                    self.swa_schedulers[k] = opt.set_scheduler(torch.optim.swa_utils.SWALR,
                                                               self.get_hparam('swa_lr', specific=k), **kwargs)
                else:
                    self.swa_schedulers[k] = torch.optim.swa_utils.SWALR(opt, self.get_hparam('swa_lr', specific=k), **kwargs)

            if k in schedulers:
                self.schedulers[k] = schedulers[k]

            elif (build_schedulers and self.get_hparam('scheduler', specific=k) is not None
                  and not self.deepspeed):

                kwargs = {'warmup': self.get_hparam('scheduler_warmup', specific=k),
                          'method': self.get_hparam('scheduler', specific=k),
                          'step_type': self.get_hparam('schedulers_steps', specific=k),
                          'cycle_momentum': True,
                          'base_momentum': self.get_hparam('cycle_base_momentum', specific=k),
                          'max_momentum': self.get_hparam('cycle_max_momentum', specific=k),
                          'patience': self.get_hparam('scheduler_patience', specific=k),
                          'factor': self.get_hparam('scheduler_factor', specific=k)}

                if type(opt) is BeamOptimizer:
                    scheduler = opt.set_scheduler(BeamScheduler, **kwargs)
                else:
                    scheduler = BeamScheduler(opt, **kwargs)

                self.schedulers[k] = scheduler

        self.refresh_optimizers_and_schedulers_pointers()

    def deepspeed_config(self, target='accelerate'):
        from ..config import deepspeed_config_generator
        config = deepspeed_config_generator(self.hparams)
        return config

    def reset_optimizers_and_schedulers(self):

        self.schedulers = {}
        self.optimizers = {}
        self.swa_schedulers = {}
        self.add_components()
        self.load_dataset(self.dataset)

    def refresh_optimizers_and_schedulers_pointers(self):
        self.optimizers_name_by_id = {id(opt): k for k, opt in self.optimizers.items()}
        self.schedulers_name_by_id = {id(sch): k for k, sch in self.schedulers.items()}
        self.schedulers_flat = self.get_flat_schedulers()
        self.optimizers_flat = self.get_flat_optimizers()

    @property
    def experiment(self):
        logger.debug(f"Fetching the experiment which is currently associated with the algorithm")
        return self._experiment

    # a setter function
    @experiment.setter
    def experiment(self, experiment):
        logger.debug(f"The algorithm is now linked to an experiment directory: {experiment.experiment_dir}")
        self.trial = experiment.trial
        self.hparams = experiment.hparams
        self.clear_experiment_properties()
        self._experiment = experiment

    def apply(self, *losses, weights=None, training=None, optimizers=None, set_to_none=True, gradient=None,
              retain_graph=None, create_graph=False, inputs=None, iteration=None, reduction=None, name=None,
              report=True):

        if training is None:
            training = self.training

        if name is None:
            name = 'loss'
        total_loss = 0

        if len(losses) == 1 and isinstance(losses[0], dict):
            losses = losses[0]
        elif len(losses) == 1:
            losses = {name: losses[0]}
        else:
            losses = {f'{name}_{i}': l for i, l in enumerate(losses)}

        if weights is None:
            weights = {k: 1 for k in losses.keys()}
        elif isinstance(weights, dict):
            pass
        else:
            weights_type = check_type(weights, check_minor=False, check_element=False)
            if weights_type.major == 'scalar':
                weights = {next(iter(losses.keys())): weights}
            else:
                weights = {f'{name}_{i}': l for i, l in enumerate(weights)}

        for k, loss in losses.items():
            n = torch.numel(loss)

            rd = self.get_hparam('reduction', specific=k) if reduction is None else reduction

            if n > 1:

                if rd == 'sum':
                    r = 1
                elif rd == 'mean':
                    r = n
                elif rd == 'mean_batch':
                    r = len(loss)
                elif rd == 'sqrt':
                    r = math.sqrt(n)
                elif rd == 'sqrt_batch':
                    r = math.sqrt(len(loss))
                else:
                    raise NotImplementedError

                loss = loss.sum()
                losses[k] = loss
                weights[k] = weights[k] / r

            total_loss = total_loss + loss * weights[k]

        if report:
            if len(losses) > 1:
                for k, l in losses.items():

                    self.report_scalar(f'{k}_s', l * weights[k])

                    if weights[k] > 1:
                        self.report_scalar(f'{k}_w', weights[k])
                    elif weights[k] == 1:
                        pass
                    elif weights[k] == 0:
                        self.report_scalar(f'{k}_w', 0)
                    else:
                        self.report_scalar(f'{k}_f', 1 / weights[k])

            self.report_scalar(name, total_loss)

        loss = total_loss
        if training:

            if self.deepspeed:
                engines = optimizers
                if engines is None:
                    engines = list(self.networks.values())
                for eng in engines:
                    eng.zero_grad()
                for eng in engines:
                    eng.backward(loss, retain_graph=retain_graph, scale_wrt_gas=True)
                for eng in engines:
                    eng.step()

            else:
                if self.amp:
                    if name is None:
                        scaler = self.scaler
                    else:
                        if name not in self.scalers:
                            self.scalers[name] = torch.cuda.amp.GradScaler()
                        scaler = self.scalers[name]

                if optimizers is None:
                    optimizers = self.optimizers_flat
                else:
                    if isinstance(optimizers, torch.optim.Optimizer) or isinstance(optimizers, BeamOptimizer):
                        optimizers = [optimizers]
                    optimizers = self.get_flat_optimizers(optimizers)

                with torch.autocast(self.autocast_device, dtype=self.mixed_precision_dtype, enabled=False):

                    it = {}

                    for k, op in optimizers.items():

                        if k not in self.optimizers_steps:
                            self.optimizers_steps[k] = 0
                        it[k] = self.optimizers_steps[k] if iteration is None else iteration
                        it[k] = (it[k] % self.get_hparam('accumulate', specific=name))

                        if not it[k]:
                            op.zero_grad(set_to_none=set_to_none)

                    if self.amp:
                        scaler.scale(loss).backward(gradient=gradient, retain_graph=retain_graph,
                                                    create_graph=create_graph, inputs=inputs)
                    elif self.accelerate:
                        self.accelerator.backward(loss, retain_graph=retain_graph)
                    elif self.deepspeed:
                        # runs backpropagation
                        op.backward(loss, retain_graph=retain_graph)

                    else:
                        loss.backward(gradient=gradient, retain_graph=retain_graph,
                                      create_graph=create_graph, inputs=inputs)

                    for k, op in optimizers.items():

                        if it[k] == self.get_hparam('accumulate', specific=name) - 1:

                            clip = self.get_hparam('clip_gradient', specific=k)
                            if clip > 0:
                                if self.amp:
                                    scaler.unscale_(op)
                                if self.accelerator is not None and self.accelerator.sync_gradients:
                                    for pg in op.param_groups:
                                        self.accelerator.clip_grad_norm_(pg['params'], clip)
                                else:
                                    for pg in op.param_groups:
                                        torch.nn.utils.clip_grad_norm_(iter(pg['params']), clip)

                            if self.amp:
                                scaler.step(op)
                            elif self.deepspeed:
                                # weight update
                                op.step()
                            else:
                                # from torch.cuda.amp import grad_scaler
                                # from accelerate import optimizer
                                op.step()

                        self.optimizers_steps[k] = self.optimizers_steps[k] + 1

        return loss

    @staticmethod
    def split_names(k):
        if '/' in k:
            k, ki = k.split('/')
        else:
            ki = None
        return k, ki

    def load_dataset(self, dataset, *args, name=None, **kwargs):
        if name is None:
            name = 'dataset'
        self.load_datasets({name: dataset}, *args, **kwargs)

    def load_datasets(self, datasets, set_epoch_length='first', batch_size_train=None, batch_size_eval=None,
                     oversample=None, weight_factor=None, expansion_size=None,timeout=0, collate_fn=None,
                     worker_init_fn=None, multiprocessing_context=None, generator=None, prefetch_factor=2,
                     dynamic=False, buffer_size=None, train_on_tail=True, probs_normalization='sum', sample_size=100000):

        batch_size_train = self.get_hparam('batch_size_train') if batch_size_train is None else batch_size_train
        batch_size_eval = self.get_hparam('batch_size_eval') if batch_size_eval is None else batch_size_eval
        oversample = (self.get_hparam('oversampling_factor') > 0) if oversample is None else oversample
        weight_factor = self.get_hparam('oversampling_factor') if weight_factor is None else weight_factor
        expansion_size = self.get_hparam('expansion_size') if expansion_size is None else expansion_size
        dynamic = self.get_hparam('dynamic_sampler') if dynamic is None else dynamic
        buffer_size = self.get_hparam('buffer_size') if buffer_size is None else buffer_size
        probs_normalization = self.get_hparam('probs_normalization') if probs_normalization is None else probs_normalization
        sample_size = self.get_hparam('sample_size') if sample_size is None else sample_size
        timeout = self.get_hparam('data_fetch_timeout') if timeout is None else timeout
        train_on_tail = self.get_hparam('train_on_tail') if train_on_tail is None else train_on_tail

        for i, (k, dataset) in enumerate(datasets.items()):

            self.datasets[k] = dataset
            if not isinstance(dataset, dict):
                subsets = dataset.indices.keys()
            else:
                subsets = dataset.keys()

            self.eval_subset = 'validation' if 'validation' in subsets else 'test'

            for s in subsets:

                if not isinstance(dataset, dict):
                    sampler = dataset.build_sampler(batch_size_eval, subset=s, persistent=False)
                    d = dataset
                else:
                    sampler = dataset[s].build_sampler(batch_size_eval, subset=None, persistent=False)
                    d = dataset[s]

                self.dataloaders[k][s] = d.build_dataloader(sampler, num_workers=self.get_hparam('cpu_workers'),
                                                               pin_memory=self.pin_memory,
                                                               timeout=timeout, collate_fn=collate_fn,
                                                               worker_init_fn=worker_init_fn,
                                                               multiprocessing_context=multiprocessing_context,
                                                               generator=generator,
                                                               prefetch_factor=prefetch_factor)
            for s in ['train', self.eval_subset]:

                if not isinstance(dataset, dict):
                    sampler = dataset.build_sampler(batch_size_train, subset=s, persistent=True, oversample=oversample,
                                                    weight_factor=weight_factor, expansion_size=expansion_size,
                                                    dynamic=dynamic, buffer_size=buffer_size,
                                                    probs_normalization=probs_normalization,
                                                    tail=train_on_tail,
                                                    sample_size=sample_size)
                    d = dataset
                else:
                    sampler = dataset[s].build_sampler(batch_size_train, subset=None, persistent=True,
                                                       oversample=oversample, weight_factor=weight_factor,
                                                       expansion_size=expansion_size,
                                                       dynamic=dynamic, buffer_size=buffer_size,
                                                       probs_normalization=probs_normalization,
                                                       tail=train_on_tail,
                                                       sample_size=sample_size)
                    d = dataset[s]

                self.persistent_dataloaders[k][s] = {'dataloader': d.build_dataloader(sampler,
                                                                  num_workers=self.get_hparam('cpu_workers'),
                                                                  pin_memory=self.pin_memory,
                                                                  timeout=timeout, collate_fn=collate_fn,
                                                                  worker_init_fn=worker_init_fn,
                                                                  multiprocessing_context=multiprocessing_context,
                                                                  generator=generator,
                                                                  prefetch_factor=prefetch_factor),
                                                  'dataset': d,
                                                  'sampler': sampler,
                                                  }
                self.persistent_dataloaders[k][s]['iterator'] = iter(self.persistent_dataloaders[k][s]['dataloader'])

            if (set_epoch_length == 'first' and i == 0) or k == set_epoch_length:
                self.set_epoch_length(dataset)

    def set_epoch_length(self, dataset):

        self.epoch_length = {'train': None, self.eval_subset: None}

        if self.get_hparam('epoch_length') is not None:
            if not isinstance(dataset, dict):
                l_train = len(dataset.indices['train'])
                l_eval = len(dataset.indices[self.eval_subset])
            else:
                l_train = len(dataset['train'])
                l_eval = len(dataset[self.eval_subset])

            self.epoch_length['train'] = int(np.round(self.get_hparam('epoch_length') * l_train / (l_train + l_eval)))
            self.epoch_length[self.eval_subset] = self.get_hparam('epoch_length') - self.epoch_length['train']

        if self.get_hparam('epoch_length_train') is not None:
            self.epoch_length['train'] = self.get_hparam('epoch_length_train')

        if self.get_hparam('epoch_length_eval') is not None:
            self.epoch_length[self.eval_subset] = self.get_hparam('epoch_length_eval')

        if self.epoch_length['train'] is None:
            self.epoch_length['train'] = len(dataset.indices['train'])

        if self.epoch_length[self.eval_subset] is None:
            self.epoch_length[self.eval_subset] = len(dataset.indices[self.eval_subset])

        if self.get_hparam('scale_epoch_by_batch_size'):
            self.epoch_length[self.eval_subset] = math.ceil(self.epoch_length[self.eval_subset] / self.batch_size_eval)
            self.epoch_length['train'] = math.ceil(self.epoch_length['train'] / self.batch_size_train)

        if self.n_epochs is None:
            self.n_epochs = self.get_hparam('total_steps') // self.epoch_length['train']

        self.set_hparam('epoch_length_train', self.epoch_length['train'])
        self.set_hparam('epoch_length_eval', self.epoch_length[self.eval_subset])

        for k, scheduler in self.schedulers_flat.items():
            if type(scheduler) is BeamScheduler:

                k, ki = self.split_names(k)

                state = self.schedulers_initial_state[k] if k in self.schedulers_initial_state else None
                if ki is not None and state is not None:
                    state = state[ki]

                scheduler.update_total_steps(epochs=self.n_epochs,
                                             steps_per_epochs=self.epoch_length['train'], initial_state=state)

    def prepare_accelerate(self):

        elements = []
        elements_origin = []
        elements_keys = []

        for k, v in self.networks.items():
            elements.append(v)
            elements_origin.append(self.networks)
            elements_keys.append((k,))
        for k, v in self.optimizers.items():
            elements.append(v)
            elements_origin.append(self.optimizers)
            elements_keys.append((k,))
        for k, v in self.schedulers.items():
            elements.append(v)
            elements_origin.append(self.schedulers)
            elements_keys.append((k,))
        # don't prepare dataloaders if you know the batch size
        # for k, v in self.persistent_dataloaders.items():
        #     for s, v in v.items():
        #         elements.append(v['dataloader'])
        #         elements_origin.append(self.persistent_dataloaders)
        #         elements_keys.append((k, s, 'dataloader'))

        modified_elements = self.accelerator.prepare(*elements)

        for e, o, kk in zip(modified_elements, elements_origin, elements_keys):
            set_item_with_tuple_key(o, kk, e)

        self.refresh_optimizers_and_schedulers_pointers()

    def prepare_deepspeed(self):

        import deepspeed
        for k, net in self.networks.items():
            opt = self.optimizers[k] if k in self.optimizers else None
            sch = self.schedulers[k] if k in self.schedulers else None

            # os.environ['LOCAL_RANK'] = str(self.rank)
            # # OMPI_COMM_WORLD_LOCAL_RANK
            # os.environ['OMPI_COMM_WORLD_RANK'] = str(self.rank)
            # # OMPI_COMM_WORLD_RANK
            # os.environ['OMPI_COMM_WORLD_SIZE'] = str(self.world_size)

            logger.critical(f"LOCAL_RANK: {os.environ['LOCAL_RANK']}")
            logger.critical(f"self.rank: {self.rank}")
            deepspeed_config = Namespace(local_rank=self.rank, device_rank=int(self.device.index),
                                         deepspeed_config=self.deepspeed_config(target='deepspeed'))

            net, opt, _, sch = deepspeed.initialize(deepspeed_config, model=net, optimizer=opt, lr_scheduler=sch)

            self.networks[k] = net

            self.optimizers[k] = opt
            if sch is not None:
                self.schedulers[k] = sch

        self.refresh_optimizers_and_schedulers_pointers()

    def get_optimizer_name(self, opt):
        i = id(opt)
        if i in self.optimizers_name_by_id:
            return self.optimizers_name_by_id[i]
        return str(i)

    def get_scheduler_name(self, sch):
        i = id(sch)
        if i in self.schedulers_name_by_id:
            return self.schedulers_name_by_id[i]
        return str(i)

    def get_flat_optimizers(self, optimizers=None):

        if isinstance(optimizers, list):
            optimizers = {self.get_optimizer_name(opt): opt for opt in optimizers}
        elif optimizers is None:
            optimizers = self.optimizers

        optimizers_flat = {}
        for k, op in optimizers.items():
            if isinstance(op, BeamOptimizer):
                for ki, opi in op.optimizers.items():
                    if len(op.optimizers) > 1:
                        optimizers_flat[f'{k}/{ki}'] = opi
                    else:
                        optimizers_flat[k] = opi
            else:
                optimizers_flat[k] = op

        return optimizers_flat

    def get_flat_schedulers(self, schedulers=None):

        if isinstance(schedulers, list):
            schedulers = {self.get_scheduler_name(sch): sch for sch in schedulers}
        elif schedulers is None:
            schedulers = self.schedulers

        schedulers_flat = {}
        for k, scheduler in schedulers.items():
            if isinstance(scheduler, MultipleScheduler):
                for ki, sch in scheduler.schedulers.items():
                    if len(scheduler.schedulers) > 1:
                        schedulers_flat[f'{k}/{ki}'] = sch
                    else:
                        schedulers_flat[k] = sch
            else:
                schedulers_flat[k] = scheduler

        return schedulers_flat

    @property
    def is_cuda(self):
        if self.device is not None and self.device.type == 'cuda':
            return True
        return False

    @property
    def ddp(self):
        return self.distributed_training and self.distributed_training_framework == 'ddp'

    def register_network(self, net, name=None):

        net = BeamNN.from_module(net, name=name, hparams=self.hparams)

        if self.device is not None:
            if self.accelerate:
                net = net.to(self.device)
            else:
                net = net.to(self.device, dtype=self.model_dtype)

        if self.ddp:

            if self.is_cuda:
                device_ids = [self.device]
            else:
                device_ids = None

            from torch.nn.parallel import DistributedDataParallel as DDP
            from types import FunctionType, MethodType

            net_ddp = DDP(net, device_ids=device_ids,
                      find_unused_parameters=self.get_hparam('find_unused_parameters', specific=name),
                      broadcast_buffers=self.get_hparam('broadcast_buffers', specific=name))

            for attr_name in dir(net):
                if attr_name not in dir(net_ddp) and not attr_name.startswith('_'):
                    # setattr(net_ddp, a, getattr(net, a))
                    attr = getattr(net.__class__, attr_name, None)
                    if isinstance(attr, property):
                        # It's a property, copy it as a property
                        setattr(net_ddp.__class__, attr_name, attr)
                    elif attr_name not in dir(net_ddp) or isinstance(getattr(net, attr_name),
                                                                     (FunctionType, MethodType)):
                        # It's not a property, but a regular attribute or a function/method; copy it directly
                        setattr(net_ddp, attr_name, getattr(net, attr_name))
            net = net_ddp

            # net = BeamDDP(net, device_ids=device_ids,
            #               find_unused_parameters=self.get_hparam('find_unused_parameters', specific=name),
            #               broadcast_buffers=self.get_hparam('broadcast_buffers', specific=name))

        if not self.accelerate and self.get_hparam('compile_network'):
            net = net.optimize('compile')

        return net

    def return_dataset(self, subset):

        if type(subset) is str or isinstance(subset, torch.utils.data.DataLoader) \
                or isinstance(subset, torch.utils.data.Dataset):
            return True
        return False

    def build_dataloaders(self, subset):
        return {k: self.build_dataloader(subset, name=k) for k in self.dataloaders.keys()}

    def build_dataloader(self, subset, name='dataset'):

        if type(subset) is str:
            dataloader = self.dataloaders[name][subset]
        elif isinstance(subset, torch.utils.data.DataLoader):
            dataloader = subset
        elif isinstance(subset, torch.utils.data.Dataset):

            dataset = subset
            sampler = dataset.build_sampler(self.get_hparam('batch_size_eval'), persistent=False)
            dataloader = dataset.build_dataloader(sampler, num_workers=self.get_hparam('cpu_workers'),
                                                  pin_memory=self.pin_memory)
        else:

            subset_type = check_type(subset)
            index = None

            if 'DataBatch' in str(type(subset)):
                dataset = UniversalDataset(subset.data, index=subset.index, label=subset.label)
            elif subset_type.minor in ['list', 'tuple']:
                dataset = UniversalDataset(*subset)
            elif subset_type.minor in ['dict']:
                dataset = UniversalDataset(**subset)
            else:
                dataset = UniversalDataset(subset)

            if index is None:
                index = len(dataset)
            sampler = UniversalBatchSampler(index, self.get_hparam('batch_size_eval'), shuffle=False,
                                            tail=True, once=True)
            dataloader = torch.utils.data.DataLoader(dataset, sampler=sampler, batch_size=None,
                                                     num_workers=self.get_hparam('cpu_workers'),
                                                     timeout=self.get_hparam('data_fetch_timeout'),
                                                     pin_memory=self.pin_memory)
        return dataloader

    def schedulers_step(self, objective=None, step_type=None):

        # when using deepspeed, the scheduler is updated implicitly by the optimizer
        if self.deepspeed:
            return

        if objective is None:
            objective = self.objective
        for k, scheduler in self.schedulers_flat.items():

            k, ki = self.split_names(k)

            if isinstance(scheduler, torch.optim.lr_scheduler._LRScheduler):
                if self.get_hparam('schedulers_steps', specific=[f'{k}/{ki}', k]) == step_type:
                    scheduler.step()
            elif isinstance(scheduler, torch.optim.lr_scheduler.ReduceLROnPlateau):
                if self.get_hparam('schedulers_steps', specific=[f'{k}/{ki}', k]) == step_type:
                    scheduler.step(objective)
            elif isinstance(scheduler, BeamScheduler):
                scheduler.step(objective, step_type=step_type)
            else:
                try:
                    scheduler.step()
                except:
                    raise Exception(f"Unknown scheduler type: {type(scheduler)}")

    def data_generator(self, subset, max_iterations=None):

        dataloader = enumerate(self.build_dataloader(subset))
        for i, (ind, label, sample) in dataloader:
            if max_iterations is not None and i >= max_iterations:
                break
            sample, label = self.to_device(sample, label)
            yield i, DataBatch(index=ind, label=label, data=sample)

    def to_device(self, *args):
        return to_device(args, self.device, half=self.half, brain=self.brain)

        # if self.accelerator is None or not self.accelerator.device_placement:
        #     return to_device(args, self.device, half=self.half, brain=self.brain)
        # else:
        #     device = 'cuda' if self.is_cuda else 'cpu'
        #     return to_device(args, device, half=self.half, brain=self.brain)

    def finite_data_generator(self, subset, length):

        dataloaders = {k: self.persistent_dataloaders[k][subset]['iterator']
                      for k in self.persistent_dataloaders.keys()}

        for i, samples in enumerate(dictionary_iterator(dataloaders)):

            for k, (ind, label, sample) in samples.items():

                sample, label = self.to_device(sample, label)
                samples[k] = DataBatch(index=ind, label=label, data=sample)

            if len(samples) == 1:
                samples = list(samples.values())[0]
            yield i, samples
            if i + 1 == length:
                break

    def preprocess_epoch(self, epoch=None, subset=None, training=True, **kwargs):
        '''
        :param aux: auxiliary data dictionary - possibly from previous epochs
        :param epoch: epoch number
        :param subset: name of dataset subset (usually train/validation/test)
        :return: None
        a placeholder for operations to execute before each epoch, e.g. shuffling/augmenting the dataset
        '''
        pass

    def train_iteration(self, sample=None, label=None, index=None, counter=None, subset=None, training=True, **kwargs):
        '''
        :param sample: the data fetched by the dataloader
        :param aux: a dictionary of auxiliary data
        :param subset: name of dataset subset (usually train/validation/test)
        :param training: train/test flag
        :return:
        loss: the loss fo this iteration
        aux: an auxiliary dictionary with all the calculated data needed for downstream computation (e.g. to calculate accuracy)
        '''
        pass

    def postprocess_epoch(self, sample=None, label=None, index=None, epoch=None, subset=None, training=True, **kwargs):
        '''
        :param epoch: epoch number
        :param subset: name of dataset subset (usually train/validation/test)
        :return: None
        a placeholder for operations to execute before each epoch, e.g. shuffling/augmenting the dataset
        '''
        pass

    @lazy_property
    def optimized_inner_train(self):
        def inner_train_with_cloned_ouptut(*args, **kwargs):
            res = self.inner_train(*args, **kwargs)
            return recursive_clone(res)

        if self.get_hparam('compile_train', False):
            return torch.compile(inner_train_with_cloned_ouptut, mode="reduce-overhead")

        return self.inner_train

    def inner_train(self, sample=None, label=None, index=None, counter=None, subset=None, training=True, **kwargs):
        raise NotImplementedError

    def iterate_epoch(self, subset, training, n):

        # if not training and self.rank > 0:
        #     return

        objective_name = self.get_hparam('objective')
        batch_size = self.batch_size_train if training else self.batch_size_eval

        with self.reporter.track_epoch(subset, batch_size=batch_size, training=training):
            if training and (n == self.n_epochs + self.swa_epochs):
                logger.warning("This is an extra epoch to calculate BN statistics. "
                               "It is not used for training so we set training=False.")
                training = False
                bu_networks = self.networks
                self.networks = self.swa_networks
                self.set_mode(training=True)

            else:
                self.set_mode(training=training)

            self.preprocess_epoch(epoch=n, training=training)
            desc = f"{subset} (epoch {n+1}/{self.n_epochs + self.swa_epochs})"


            data_generator = self.finite_data_generator(subset, self.epoch_length[subset])
            for i, samples in self.reporter.iterate(data_generator,
                                  enable=self.enable_tqdm, notebook=(not self.ddp and self.is_notebook),
                                  threshold=self.get_hparam('tqdm_threshold'), stats_period=self.get_hparam('tqdm_stats'),
                                  desc=desc, total=self.epoch_length[subset]):

                if type(samples) is DataBatch:
                    ind = samples.index
                    label = samples.label
                    sample = samples.data
                else:
                    sample = samples
                    ind = None
                    label = None

                with torch.autocast(self.autocast_device, dtype=self.mixed_precision_dtype,
                                    enabled=self.amp):
                    self.train_iteration(sample=sample, counter=i, training=training, index=ind, label=label)

                    objective = self.reporter.get_scalar(objective_name, subset=subset, aggregate=False, index=-1)

                    if training and n < self.n_epochs:
                        self.schedulers_step(objective, step_type='iteration')

                    if self.amp and training:
                        if self.scaler._scale is not None:
                            self.scaler.update()
                        for k, scaler in self.scalers.items():
                            if scaler._scale is not None:
                                scaler.update()

            self.postprocess_epoch(sample=sample, index=ind, label=label, epoch=n, training=training)

            if n == self.n_epochs + self.swa_epochs:
                self.set_mode(training=False)
                self.networks = bu_networks

    def preprocess_inference(self, subset=None, predicting=False, **argv):
        '''
        :param aux: auxiliary data dictionary - possibly from previous epochs
        :param subset: name of dataset subset (usually train/validation/test)
        :return: None
        a placeholder for operations to execute before each epoch, e.g. shuffling/augmenting the dataset
        '''
        pass

    def inference_iteration(self, sample=None, label=None, index=None, subset=None, predicting=False, **kwargs):
        '''
        :param sample: the data fetched by the dataloader
        :param aux: a dictionary of auxiliary data
        :param subset: name of dataset subset (usually train/validation/test)
        :return:
        loss: the loss fo this iteration
        aux: an auxiliary dictionary with all the calculated data needed for downstream computation (e.g. to calculate accuracy)
        '''
        self.train_iteration(sample=sample, label=label, index=index, subset=subset, counter=0, training=False, **kwargs)
        return {}

    def postprocess_inference(self, sample=None, label=None, index=None, subset=None, predicting=False, **kwargs):
        '''
        :param subset: name of dataset subset (usually train/validation/test)
        :return: None
        a placeholder for operations to execute before each epoch, e.g. shuffling/augmenting the dataset
        '''
        pass

    def calculate_objective(self):
        '''
        This function calculates the optimization non-differentiable objective. It is used for hyperparameter optimization
        and for ReduceOnPlateau scheduling. It is also responsible for tracking the best checkpoint
        '''

        self.best_objective = self.reporter.best_objective
        self.best_epoch = self.reporter.best_epoch
        self.objective = self.reporter.objective
        self.best_state = self.reporter.best_state

        return self.objective

    def report(self, objective, epoch=None):
        '''
        Use this function to report results to hyperparameter optimization frameworks
        also you can add key 'objective' to the results dictionary to report the final scores.
        '''

        if self.hpo == 'tune':

            if self.get_hparam('objective') is not None:
                metrics = {self.get_hparam('objective'): objective}
            else:
                metrics = {'objective': objective}
            from ray import train
            train.report(metrics)
        elif self.hpo == 'optuna':
            import optuna
            self.trial.report(objective, epoch)
            self.trial.set_user_attr('best_value', self.best_objective)
            self.trial.set_user_attr('best_epoch', self.best_epoch)
            if self.trial.should_prune():
                raise optuna.TrialPruned()

            train_timeout = self.get_hparam('train-timeout')
            if train_timeout is not None and 0 < train_timeout < self.elapsed_time:
                raise optuna.exceptions.OptunaError(f"Trial timed out after {self.get_hparam('train-timeout')} seconds.")

    @lazy_property
    def optimization_mode(self):
        objective_mode = self.get_hparam('objective_mode')
        objective_name = self.get_hparam('objective')
        return self.get_optimization_mode(objective_mode, objective_name)

    @staticmethod
    def get_optimization_mode(mode, objective_name):
        if mode is not None:
            return mode
        if any(n in objective_name.lower() for n in ['loss', 'error', 'mse']):
            return 'min'
        return 'max'

    def early_stopping(self, epoch=None):
        '''
        Use this function to early stop your model based on the results or any other metric in the algorithm class
        '''

        if self.rank > 0:
            return False

        train_timeout = self.get_hparam('train-timeout')
        if train_timeout is not None and 0 < train_timeout < self.elapsed_time:
            logger.info(f"Stopping training at epoch {self.epoch} - timeout {self.get_hparam('train-timeout')}")
            return True

        stop_at = self.get_hparam('stop_at')
        early_stopping_patience = self.get_hparam('early_stopping_patience')
        if self.objective is None and stop_at is not None:
            logger.warning("Early stopping is enabled (stop_at is not None) but no objective is defined. "
                           "set objective in the hparams")
            return False
        if self.objective is None and early_stopping_patience is not None:
            logger.warning("Early stopping is enabled (early_stopping_patience is not None) "
                           "but no objective is defined. set objective in the hparams")
            return False

        if stop_at is not None:
            if self.best_objective is not None:

                if self.optimization_mode == 'max':
                    res = self.best_objective > stop_at
                    if res:
                        logger.info(f"Stopping training at {self.best_objective} > {stop_at}")
                else:
                    res = self.best_objective < stop_at
                    if res:
                        logger.info(f"Stopping training at {self.best_objective} < {stop_at}")
                return res

        if early_stopping_patience is not None and early_stopping_patience > 0:
            res = self.epoch - self.best_epoch > early_stopping_patience
            if res:
                logger.info(f"Stopping training at epoch {self.epoch} - best epoch {self.best_epoch} > {early_stopping_patience}")
            return res

        return False

    def __call__(self, subset, dataset_name='dataset', predicting=False, enable_tqdm=None, max_iterations=None,
                 head=None, eval_mode=True, return_dataset=None, **kwargs):

        self.set_reporter(BeamReport(objective=self.get_hparam('objective'),
                                     objective_mode=self.optimization_mode))

        with torch.no_grad():

            self.set_mode(training=not eval_mode)
            transforms = []
            index = []

            desc = subset if type(subset) is str else ('predict' if predicting else 'evaluate')

            if enable_tqdm is None:
                enable_tqdm = self.enable_tqdm

            if return_dataset is None:
                if predicting:
                    return_dataset = self.return_dataset(subset)
                    if not return_dataset:
                        logger.warning("Predicting: the inferred return type will be DataBatch and results statistics "
                                       "will be omitted. To avoid this behavior please provide a dataset or specify "
                                       "return_dataset=True")
                else:
                    return_dataset = True

            dataloader = self.build_dataloader(subset, name=dataset_name)
            dataset = dataloader.dataset

            batch_size = self.batch_size_eval
            if head is not None:
                max_iterations = math.ceil(head / batch_size)

            with (self.reporter.track_epoch(desc, batch_size=batch_size, training=not eval_mode)):
                self.preprocess_inference(subset=subset, predicting=predicting, dataset=dataset, **kwargs)
                data_generator = self.data_generator(dataloader, max_iterations=max_iterations)
                total_iterations = len(dataloader) if max_iterations is None else min(len(dataloader), max_iterations)
                for i, (ind, label, sample) in self.reporter.iterate(data_generator, enable=enable_tqdm,
                                      threshold=self.get_hparam('tqdm_threshold'), stats_period=self.get_hparam('tqdm_stats'),
                                      notebook=(not self.ddp and self.is_notebook), desc=desc, total=total_iterations):
                    transform = self.inference_iteration(sample=sample, subset=subset, predicting=predicting,
                                                         label=label, index=ind, **kwargs)
                    transforms.append(transform)
                    index.append(ind)

                index = torch.cat(index)
                transforms = recursive_concatenate(transforms)

                self.postprocess_inference(sample=sample, index=ind, transforms=transforms, label=label,
                                                     subset=subset, dataset=dataset, predicting=predicting, **kwargs)

            if return_dataset:
                dataset = UniversalDataset(transforms, index=index)
                dataset.set_statistics(self.reporter.data)
            else:
                dataset = DataBatch(index=index, data=transforms, label=None)

        return dataset

    def training_closure(self, *args, **kwargs):
        pass

    def __iter__(self):

        self.t0 = timer()

        n_epochs = self.n_epochs + self.swa_epochs + int(self.swa_epochs > 0)
        self.refresh_optimizers_and_schedulers_pointers()

        if self.accelerate:
            self.prepare_accelerate()
        elif self.deepspeed:
            self.prepare_deepspeed()

        epoch_start = 0 if self.get_hparam("restart_epochs_count", default=True) else self.epoch
        self.set_train_reporter(first_epoch=epoch_start, n_epochs=n_epochs)

        for i in range(epoch_start, n_epochs):

            self.reporter.reset_epoch(i, total_epochs=self.epoch + 1)

            self.iterate_epoch(subset='train', training=True, n=i)
            with torch.no_grad():
                self.iterate_epoch(subset=self.eval_subset, training=False, n=i)

            # add learning rate and momentum of schedulers_steps
            for k, scheduler in self.schedulers_flat.items():

                if self.accelerate:
                    lr = scheduler.optimizers[0].param_groups[0]['lr']
                else:
                    lr = scheduler.optimizer.param_groups[0]['lr']

                self.report_scalar(f'lr_{k}', lr, subset='train')
                if type(scheduler) is BeamScheduler and scheduler.method in ['one_cycle']:
                    self.report_scalar(f'momentum_{k}', scheduler.get_current_state()['momentum'], subset='train')

            self.epoch += 1
            
            objective = self.calculate_objective()

            if self.get_hparam('objective_to_report') == 'last':
                report_objective = objective
            elif self.get_hparam('objective_to_report') == 'best':
                report_objective = self.best_objective
            else:
                raise Exception(f"Unknown objective_to_report: {self.get_hparam('objective_to_report')} "
                                f"should be [last|best]")
            self.report(report_objective, i)

            if i+1 == self.n_epochs and self.swa_epochs > 0:
                logger.warning("Switching to SWA training")

            if i+1 >= self.n_epochs and self.swa_epochs > 0:
                for k, swa_model in self.swa_networks.items():
                    swa_model.update_parameters(self.networks[k])
                for k, sch in self.swa_schedulers.items():
                    sch.step()

                    lr = sch.optimizer.param_groups[0]['lr']
                    self.report_scalar(f'swalr_{k}', lr, subset='train')
            else:
                self.schedulers_step(objective=objective, step_type='epoch')

            yield self.reporter

            if self.early_stopping(i):
                return

    def set_mode(self, training=True):

        for net in self.networks.values():

            if training:
                net.train()
            else:
                net.eval()

        for dataset in self.datasets.values():
            if hasattr(dataset, 'train'):
                if training:
                    dataset.train()
                else:
                    dataset.eval()

        self.training = training

    def state_dict(self):
        return self.save_checkpoint(networks=True, optimizers=False, schedulers=False,
                             processors=True, scaler=False, scalers=False, swa_schedulers=False, swa_networks=False,
                             aux=None, pickle_model=False, hparams=True)

    def load_state(self, path):
        self.load_checkpoint(path, networks=True, optimizers=False, schedulers=False, processors=True,
                        scaler=False, scalers=False, swa_schedulers=False, swa_networks=False, hparams=False)

    def save_checkpoint(self, path=None, networks=True, optimizers=True, schedulers=True,
                        processors=True, scaler=True, scalers=True, swa_schedulers=True, swa_networks=True,
                        hparams=True, aux=None, pickle_model=False):

        path = beam_path(path)
        state = {'aux': aux, 'epoch': self.epoch, 'best_objective': self.best_objective, 'best_epoch': self.best_epoch,}
        wrapper = copy.deepcopy if path is None else (lambda x: x)

        if hparams:
            state['hparams'] = self.hparams.namespace

        if not self.deepspeed:

            for k, net in filter_dict(self.networks, networks).items():
                state[f"{k}_parameters"] = wrapper(net.state_dict())
                if pickle_model:
                    state[f"{k}_model"] = net

            for k, optimizer in filter_dict(self.optimizers, optimizers).items():
                state[f"{k}_optimizer"] = wrapper(optimizer.state_dict())

            for k, scheduler in filter_dict(self.schedulers, schedulers).items():
                state[f"{k}_scheduler"] = wrapper(scheduler.state_dict())

            for k, processor in filter_dict(self.processors, processors).items():
                state[f"{k}_processor"] = wrapper(processor.state_dict())

            for k, swa_scheduler in filter_dict(self.swa_schedulers, swa_schedulers).items():
                state[f"{k}_swa_scheduler"] = wrapper(swa_scheduler.state_dict())

            for k, swa_network in filter_dict(self.swa_networks, swa_networks).items():
                state[f"{k}_swa_network"] = wrapper(swa_network.state_dict())

            if scaler:
                state['scaler'] = self.scaler.state_dict() if self.scaler is not None else None

            if scalers:
                state['scalers'] = {k: scaler.state_dict()
                                    if scaler is not None else None for k, scaler in self.scalers.items()}

            if path is not None:
                path.write(state, ext=self.file_format)
            else:
                return state

        else:

            def save_models(path):

                path.mkdir()
                for k, net in filter_dict(self.networks, networks).items():
                    net.save_checkpoint(str(path), client_state=state)

            if path.is_local:
                save_models(path)

            with local_copy(path) as path:
                save_models(path)

    @property
    def file_format(self):
        if self.get_hparam('safetensors', default=False):
            return '.safetensors'
        return '.pt'

    def load_checkpoint(self, path_or_state, strict=True, networks=True, optimizers=True, schedulers=True, hparams=True,
                        processors=True, scaler=True, scalers=True, swa_schedulers=True, swa_networks=True, load_epoch=True):

        path_or_state = beam_path(path_or_state)

        if self.deepspeed:

            def load_models(path):

                client_state = {}
                for k, net in filter_dict(self.networks, networks).items():
                    _, client_state = net.load_checkpoint(str(path), load_module_strict=strict,
                                                    load_optimizer_states=optimizers,
                                                    load_lr_scheduler_states=schedulers,
                                                    load_module_only=False)

                return client_state

            if path_or_state.is_local:
                state = load_models(path_or_state)
            else:
                with local_copy(path_or_state) as path:
                    state = load_models(path)

        else:

            if hasattr(path_or_state, 'read') and hasattr(path_or_state, 'as_uri') :
                logger.info(f"Loading network state from: {path_or_state}")
                state = path_or_state.read(ext=self.file_format, map_location=self.device)
            else:
                state = path_or_state

            for k, net in filter_dict(self.networks, networks).items():

                if f"{k}_parameters" in state.keys():
                    s = state[f"{k}_parameters"]

                    if not self.ddp:
                        torch.nn.modules.utils.consume_prefix_in_state_dict_if_present(s, 'module.')

                    net.load_state_dict(s, strict=strict)
                else:
                    logger.warning(f"Network {k} is missing from the state-dict")

            for k, net in filter_dict(self.swa_networks, swa_networks).items():

                if f"{k}_swa_network" in state.keys():
                    s = state[f"{k}_swa_network"]

                    if not self.ddp:
                        torch.nn.modules.utils.consume_prefix_in_state_dict_if_present(s, 'module.')

                    net.load_state_dict(s, strict=strict)
                else:
                    logger.warning(f"SWA Network {k} is missing from the state-dict")

            for k, optimizer in filter_dict(self.optimizers, optimizers).items():
                if f"{k}_optimizer" in state.keys():
                    optimizer.load_state_dict(state[f"{k}_optimizer"])
                else:
                    logger.warning(f"Optimizer {k} is missing from the state-dict")

            for k, processor in filter_dict(self.processors, processors).items():
                if f"{k}_processor" in state.keys():
                    processor.load_state_dict(state[f"{k}_processor"])
                else:
                    logger.warning(f"Processor {k} is missing from the state-dict")

            for k, scheduler in filter_dict(self.schedulers, schedulers).items():
                if f"{k}_scheduler" in state.keys():
                    self.schedulers_initial_state[k] = state[f"{k}_scheduler"]
                    try:
                        scheduler.load_state_dict(state[f"{k}_scheduler"])
                    except AttributeError:
                        logger.warning("Tries to load scheduler which requires dataset info: please load dataset first")
                else:
                    logger.warning(f"Scheduler {k} is missing from the state-dict")

            for k, swa_scheduler in filter_dict(self.swa_schedulers, swa_schedulers).items():
                if f"{k}_swa_scheduler" in state.keys():
                    swa_scheduler.load_state_dict(state[f"{k}_swa_scheduler"])
                else:
                    logger.warning(f"SWA Scheduler {k} is missing from the state-dict")

            if scaler:
                if self.scaler is not None and 'scaler' in state.keys():
                    self.scaler.load_state_dict(state["scaler"])

            if "scalers" in state.keys():
                for k, s in filter_dict(self.scalers, scalers).items():
                    if k in state["scalers"]:
                        self.scalers[k].load_state_dict(state["scalers"][k])

        if load_epoch and 'epoch' in state.keys():
            self.epoch = state['epoch']

        if hparams and 'hparams' in state.keys():
            self.hparams = state['hparams']

        return state.pop('aux', None)

    def set_auxiliary_state(self, aux):
        pass

    def optimize(self, method='compile', networks=True, **kwargs):

        for k, net in filter_dict(self.networks, networks).items():

            logger.warning(f"Optimizing model: {k} with torch compile")
            if self.accelerator is None:
                self.networks[k] = net.optimize(method, **kwargs)
            else:
                assert method=='compile', "Accelerate does not support other optimization methods than compile"
                self.networks[k] = self.accelerator.prepare_model(net, evaluate=True)

    def fit(self, dataset=None, dataloaders=None, timeout=0, collate_fn=None,
                   worker_init_fn=None, multiprocessing_context=None, generator=None, prefetch_factor=2, **kwargs):
        '''
        For training purposes
        '''

        def algorithm_generator_single(experiment, *args, **kwargs):

            if dataset is not None:
                self.load_dataset(dataset=dataset, dataloaders=dataloaders, timeout=0, collate_fn=None,
                                  worker_init_fn=None, multiprocessing_context=None, generator=None, prefetch_factor=2)

            return self

        if self.get_hparam('n_gpus') == 1:
            algorithm_generator = algorithm_generator_single
        else:
            raise NotImplementedError("To continue training in parallel mode: please re-run experiment() with "
                                      "your own algorithm generator and a new dataset")

        assert self._experiment is not None, "No experiment is linked with the current algorithm"

        return self._experiment(algorithm_generator, **kwargs)

    def evaluate(self, *args, **kwargs):
        '''
        For validation and test purposes (when labels are known)
        '''
        return self(*args, predicting=False, **kwargs)

    def score(self, *args, **kwargs):
        '''
        For validation and test purposes (when labels are known)
        '''
        return self.evaluate(*args, predicting=False, **kwargs)

    def predict(self, dataset, *args, lazy=False, kpi=True, **kwargs):
        '''
        For real data purposes (when labels are unknown)
        '''
        if lazy:
            return TransformedDataset(dataset, self, *args, **kwargs)

        if not kpi:
            return self(dataset, *args, predicting=True, **kwargs)

        @beam_kpi(BeamResult)
        def predict_wrapper(sample, algorithm=self, **kwargs):
            return algorithm(sample, predicting=True, **kwargs)

        return predict_wrapper(dataset, algorithm=self, *args, **kwargs)



import numpy as np
import torch
from ..algorithm import NeuralAlgorithm
from ..logger import beam_logger as logger
from ..utils import as_numpy
from .beam_similarity import BeamSimilarity


class BeamSSL(NeuralAlgorithm):

    def __init__(self, hparams, networks=None, optimizers=None, schedulers=None, dataset=None, labeled_dataset=None):

        if networks is None:
            networks = {}

        encoder = self.generate_encoder()
        if encoder is not None:
            networks['encoder'] = encoder

        self.logger = logger

        super().__init__(hparams, networks=networks, optimizers=optimizers, schedulers=schedulers)

        if labeled_dataset is None:
            labeled_dataset = self.generate_labeled_set()
        self.labeled_dataset = labeled_dataset

        self.index_train_labeled = np.array(self.labeled_dataset.indices['train'])
        self.index_test_labeled = np.array(self.labeled_dataset.indices['test'])
        self.sim = None

    def generate_labeled_set(self, *args, pretrained=None, **kwargs):
        """
        This function should be overridden by the child class. Its purpose is to generate a labeled test-set for the
        evaluation of the downstream task.
        @return: UniversalDataset
        """
        return None

    def generate_encoder(self, *args, pretrained=None, **kwargs):
        """
        This function should be overridden by the child class. Its purpose is to generate a fresh
        (untrained or pretrained) encoder.
        @param pretrained:
        @return: nn.Module
        """
        return None

    @property
    def p_dim(self):
        raise NotImplementedError

    @property
    def h_dim(self):
        raise NotImplementedError

    def preprocess_inference(self, results=None, augmentations=0, dataset=None, **kwargs):

            if augmentations > 0 and dataset is not None:
                results['aux']['org_n_augmentations'] = dataset.n_augmentations
                dataset.n_augmentations = augmentations

            return results

    def postprocess_inference(self, sample=None, results=None, subset=None, dataset=None, **kwargs):

        if 'aux' in results and 'org_n_augmentations' in results['aux'] and dataset is not None:
            dataset.n_augmentations = results['aux']['org_n_augmentations']

        return results

    def evaluate_downstream_task(self, z, y):

        train_data = lgb.Dataset(z[self.index_train_labeled], label=y[self.index_train_labeled])
        validation_data = lgb.Dataset(z[self.index_test_labeled], label=y[self.index_test_labeled])

        if self.hparams.lgb_device is None:
            device = None if 'cpu' == self.device.type else self.device.index
        else:
            device = self.hparams.lgb_device

        num_round = self.hparams.lgb_rounds
        param = {'objective': 'multiclass',
                 'num_leaves': self.hparams.lgb_num_leaves,
                 'max_depth': self.hparams.lgb_max_depth,
                 'gpu_device_id': device,
                 'verbosity': -1,
                 'metric': ['multi_error', 'multiclass'],
                 'num_class': np.max(y) + 1}

        return lgb.train(param, train_data, num_round, valid_sets=[validation_data], verbose_eval=self.hparams.verbose_lgb)

    def postprocess_epoch(self, results=None, training=None, epoch=None, **kwargs):

        if not training and not epoch % 1:

            self.logger.info("Evaluating the downstream task")
            features = self.evaluate(self.labeled_dataset, projection=False, prediction=False, augmentations=0)
            z = as_numpy(features.values['h'])
            y = as_numpy(features.values['y'])

            bst = self.evaluate_downstream_task(z, y)

            results['scalar']['encoder_acc'] = 1 - bst.best_score['valid_0']['multi_error']
            results['scalar']['encoder_loss'] = bst.best_score['valid_0']['multi_logloss']

            if 'z' in features.values:

                z = as_numpy(features.values['z'])
                bst = self.evaluate_downstream_task(z, y)

                results['scalar']['projection_acc'] = 1 - bst.best_score['valid_0']['multi_error']
                results['scalar']['projection_loss'] = bst.best_score['valid_0']['multi_logloss']

                if 'p' in features.values:

                    z = as_numpy(features.values['p'])
                    bst = self.evaluate_downstream_task(z, y)

                    results['scalar']['prediction_acc'] = 1 - bst.best_score['valid_0']['multi_error']
                    results['scalar']['prediction_loss'] = bst.best_score['valid_0']['multi_logloss']

        return results

    def build_similarity(self, add_sets=None, train_sets=None, metric='l2', training_device=None, inference_device=None,
                         ram_footprint=2 ** 8 * int(1e9), gpu_footprint=24 * int(1e9), exact=False, nlists=None,
                         M=None, latent_variable='h', projection=False, prediction=False):

        device = self.device
        device = device.type if 'cpu' == device.type else device.index

        if training_device is None:
            training_device = device
        if inference_device is None:
            inference_device = device

        if add_sets is None:
            add_sets = ['train', self.eval_subset, self.labeled_dataset]
        if train_sets is None:
            train_sets = add_sets

        d = self.h_dim

        expected_population = 0
        add_dataloaders = {}
        for subset in add_sets:
            dataloader = self.build_dataloader(subset)
            expected_population += len(dataloader.dataset)
            add_dataloaders[id(subset)] = dataloader

        train_population = 0
        train_dataloaders = {}
        for subset in train_sets:
            dataloader = self.build_dataloader(subset)
            train_population += len(dataloader.dataset)
            train_dataloaders[id(subset)] = dataloader

        self.sim = BeamSimilarity(d=d, expected_population=expected_population,
                                  metric=metric, training_device=training_device, inference_device=inference_device,
                                  ram_footprint=ram_footprint, gpu_footprint=gpu_footprint, exact=exact,
                                  nlists=nlists, M=M, reducer='umap')

        h = []
        for i, dataloader in add_dataloaders.items():
            predictions = self.predict(dataloader, prediction=prediction, projection=projection,
                                       add_to_sim=True, latent_variable=latent_variable)
            if i in train_dataloaders:
                h.append(predictions.data[latent_variable])

        for i, dataloader in train_dataloaders.items():
            if i not in add_dataloaders:
                predictions = self.predict(dataloader, prediction=prediction, projection=projection,
                                       add_to_sim=False, latent_variable=latent_variable)

                h.append(predictions.data[latent_variable])

        h = torch.cat(h)
        self.sim.train(h)

        return self.sim

    def inference_iteration(self, sample=None, results=None, subset=None, predicting=True, similarity=0,
                            projection=True, prediction=True, augmentations=0, add_to_sim=False, latent_variable='h',
                            **kwargs):

        data = {}
        if isinstance(sample, dict):
            x = sample['x']
            if 'y' in sample:
                data['y'] = sample['y']
        else:
            x = sample

        networks = self.networks

        # b = len(x)
        # if b < self.batch_size_eval:
        #     x = torch.cat([x, torch.zeros((self.batch_size_eval-b, *x.shape[1:]), device=x.device, dtype=x.dtype)])

        h = networks['encoder'](x)

        # if b < self.batch_size_eval:
        #     h = h[:b]

        data['h'] = h

        if 'projection' in networks and projection:
            z = networks['projection'](h)
            data['z'] = z

        if 'prediction' in networks and prediction:
            p = networks['prediction'](z)
            data['p'] = p

        if isinstance(sample, dict) and 'augmentations' in sample and augmentations:
            representations = []
            for a in sample['augmentations']:
                representations.append(networks['encoder'](a))

            representations = torch.stack(representations)

            mu = representations.mean(dim=0)
            std = representations.std(dim=0)

            results['scalar']['mu'] = mu
            results['scalar']['std'] = std

        if add_to_sim:
            if self.sim is not None:
                self.sim.add(data[latent_variable], train=False)
            else:
                logger.error("Please build similarity object first before adding indices. Use alg.build_similarity()")

        if similarity > 0:
            if self.sim is not None:
                similarities = self.sim.search(data[latent_variable], k=similarity)

                data['similarities_index'] = similarities.index
                data['similarities_distance'] = similarities.distance

            else:
                logger.error("Please build and train similarity object first before calculating similarities. "
                             "Use alg.build_similarity()")

        return data, results



import json
from typing import Optional, Any
import pandas as pd
import numpy as np

from ..logger import beam_logger as logger
from .core import BeamLLM, CompletionObject
from pydantic import Field, PrivateAttr
from ..path import beam_key

from ..utils import lazy_property


class OpenAIBase(BeamLLM):

    api_key: Optional[str] = Field(None)
    api_base: Optional[str] = Field(None)
    organization: Optional[str] = Field(None)

    def __init__(self, api_key=None, api_base=None, organization=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.api_key = api_key
        self.api_base = api_base
        self.organization = organization

    @lazy_property
    def client(self):
        from openai import OpenAI
        return OpenAI(organization=self.organization, api_key=self.api_key, base_url=self.api_base)

    def update_usage(self, response):

        if 'usage' in response:
            response = response['usage']

            self.usage["prompt_tokens"] += response["prompt_tokens"]
            self.usage["completion_tokens"] += response["completion_tokens"]
            self.usage["total_tokens"] += response["prompt_tokens"] + response["completion_tokens"]

    # def sync_openai(self):
    #     openai.api_key = self.api_key
    #     openai.api_base = self.api_base
    #     openai.organization = self.organization

    def _completion(self, prompt=None, **kwargs):
        # self.sync_openai()
        res = self.client.completions.create(model=self.model, prompt=prompt, **kwargs)
        return CompletionObject(prompt=prompt, kwargs=kwargs, response=res)

    def _chat_completion(self, messages=None, **kwargs):
        # self.sync_openai()
        res = self.client.chat.completions.create(model=self.model, messages=messages, **kwargs)
        return CompletionObject(prompt=messages, kwargs=kwargs, response=res)

    def verify_response(self, res):
        stream = res.stream
        res = res.response

        if not hasattr(res.choices[0], 'finish_reason'):
            return False

        finish_reason = res.choices[0].finish_reason
        if finish_reason != 'stop' and not stream:
            logger.warning(f"finish_reason is {finish_reason}")

        return True

    def extract_text(self, res):

        stream = res.stream
        res = res.response

        if not self.is_chat:
            res = res.choices[0].text
        else:
            if not stream:
                res = res.choices[0].message.content
            else:
                res = res.choices[0].delta.content
        return res

    def openai_format(self, res):

        res = res.response
        return res


class OpenAILLM(OpenAIBase):

    _models: Any = PrivateAttr(default=None)

    def __init__(self, model='gpt-3.5-turbo', api_key=None, organization=None, *args, **kwargs):

        api_key = beam_key('OPENAI_API_KEY', api_key)

        kwargs['scheme'] = 'openai'
        super().__init__(api_key=api_key, api_base='https://api.openai.com/v1',
                         organization=organization, *args, **kwargs)

        self.model = model
        self._models = None

    @property
    def is_chat(self):
        instruct_models = ['gpt-3.5-turbo-instruct', 'babbage-002', 'davinci-002']
        if self.model in instruct_models:
            return False
        return True

    def file_list(self):
        import openai
        return openai.File.list()

    def build_dataset(self, data=None, question=None, answer=None, path=None) -> object:
        """
        Build a dataset for training a model
        :param data: dataframe with prompt and completion columns
        :param question: list of questions
        :param answer: list of answers
        :param path: path to save the dataset
        :return: path to the dataset
        """
        if data is None:
            data = pd.DataFrame(data={'prompt': question, 'completion': answer})

        records = data.to_dict(orient='records')

        if path is None:
            logger.warning('No path provided, using default path: dataset.jsonl')
            path = 'dataset.jsonl'

        # Open a file for writing
        with open(path, 'w') as outfile:
            # Write each data item to the file as a separate line
            for item in records:
                json.dump(item, outfile)
                outfile.write('\n')

        return path

    def retrieve(self, model=None):
        import openai
        if model is None:
            model = self.model
        return openai.Engine.retrieve(id=model)

    @property
    def models(self):
        if self._models is None:
            import openai
            models = openai.models.list()
            models = {m.id: m for m in models.data}
            self._models = models
        return self._models

    def embedding(self, text, model=None):
        if model is None:
            model = self.model
        import openai
        response = openai.Engine(model).embedding(input=text, model=model)
        embedding = np.array(response.data[1]['embedding'])
        return embedding




from typing import Union

from dataclasses import make_dataclass
from ..core import Processor
from ..utils import lazy_property
from kubernetes import client, watch
from kubernetes.client import Configuration, RbacAuthorizationV1Api, V1DeleteOptions
from kubernetes.client.rest import ApiException
from kubernetes.stream import stream
from ..logger import beam_logger as logger
from .units import K8SUnits
import time
import json


class BeamK8S(Processor):  # processor is another class and the BeamK8S inherits the method of processor
    """BeamK8S is a class  that  provides a simple interface to the Kubernetes API."""

    def __init__(self, api_url=None, api_token=None, namespace=None,
                 project_name=None, use_scc=None, scc_name=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.api_token = api_token
        self.api_url = api_url
        self.project_name = project_name
        self.namespace = namespace
        self.use_scc = use_scc
        self.scc_name = scc_name

    @lazy_property
    def core_v1_api(self):
        return client.CoreV1Api(self.api_client)

    @lazy_property
    def api_client(self):
        return client.ApiClient(self.configuration)

    @lazy_property
    def apps_v1_api(self):
        return client.AppsV1Api(self.api_client)

    @lazy_property
    def configuration(self):
        configuration = Configuration()
        configuration.host = self.api_url
        configuration.verify_ssl = False  # Depends on your SSL setup
        configuration.debug = False
        configuration.api_key = {
            'authorization': f"Bearer {self.api_token}"
        }
        return configuration

    @lazy_property
    def dyn_client(self):
        from openshift.dynamic import DynamicClient
        # Ensure the api_client is initialized before creating the DynamicClient
        return DynamicClient(self.api_client)

    @lazy_property
    def rbac_api(self):
        return RbacAuthorizationV1Api(self.api_client)

    def create_project(self, project_name):
        try:
            # Attempt to get the project to see if it already exists
            self.dyn_client.resources.get(api_version='project.openshift.io/v1', kind='Project').get(name=project_name)
            logger.info(f"Project '{project_name}' already exists.")
        except ApiException as e:
            if e.status == 404:  # Project does not exist, create it
                project_request = {
                    "kind": "ProjectRequest",
                    "apiVersion": "project.openshift.io/v1",
                    "metadata": {"name": project_name}
                }
                self.dyn_client.resources.get(api_version='project.openshift.io/v1',
                                              kind='ProjectRequest').create(body=project_request)
                logger.info(f"Project '{project_name}' created successfully.")
            else:
                logger.error(f"Failed to check or create project '{project_name}': {e}")

    def create_service_account(self, name, namespace=None):
        namespace = namespace or self.namespace

        # Attempt to read the service account, create it if it does not exist
        try:
            self.core_v1_api.read_namespaced_service_account(name, namespace)
            logger.info(f"Service Account {name} already exists in namespace {namespace}.")
        except ApiException as e:
            if e.status == 404:  # Not Found
                metadata = client.V1ObjectMeta(name=name)
                service_account = client.V1ServiceAccount(api_version="v1", kind="ServiceAccount", metadata=metadata)
                self.core_v1_api.create_namespaced_service_account(namespace=namespace, body=service_account)
                logger.info(f"Service Account {name} created in namespace {namespace}.")
            else:
                logger.error(f"Failed to check or create Service Account {name} in namespace {namespace}: {e}")
                raise

    def add_scc_to_service_account(self, service_account_name, namespace, scc_name):
        scc = self.dyn_client.resources.get(api_version='security.openshift.io/v1', kind='SecurityContextConstraints')
        scc_obj = scc.get(name=scc_name)
        user_name = f"system:serviceaccount:{namespace}:{service_account_name}"
        if user_name not in scc_obj.users:
            scc_obj.users.append(user_name)
            scc.patch(body=scc_obj, name=scc_name, content_type='application/merge-patch+json')

    @staticmethod
    def create_container(image_name, deployment_name=None, project_name=None, ports=None, pvc_mounts=None,
                         cpu_requests=None, cpu_limits=None, memory_requests=None,
                         memory_limits=None, gpu_requests=None, memory_storage_configs=None,
                         gpu_limits=None, security_context_config=None, security_context=None,
                         entrypoint_args=None, entrypoint_envs=None):
        container_name = f"{project_name}-{deployment_name}-container" \
            if project_name and deployment_name else "default-container"

        # Preparing environment variables from entrypoint_args and envs
        env_vars = []
        entrypoint_args = entrypoint_args or []
        entrypoint_envs = entrypoint_envs or {}

        for arg in entrypoint_args:
            env_vars.append(client.V1EnvVar(name=f"ARG_{arg}", value=str(arg)))
        for key, value in entrypoint_envs.items():
            env_vars.append(client.V1EnvVar(name=key, value=str(value)))

        # Preparing volume mounts
        volume_mounts = []
        if pvc_mounts:
            for mount in pvc_mounts:
                volume_mounts.append(client.V1VolumeMount(
                    name=mount['pvc_name'],  # Using PVC name as the volume name
                    mount_path=mount['mount_path']  # Mount path specified in pvc_mounts
                ))

            # Handling memory_storage_configs - this is the only addition
        if memory_storage_configs:
            for mem_storage in memory_storage_configs:
                volume_mounts.append(client.V1VolumeMount(
                    name=mem_storage.name,
                    mount_path=mem_storage.mount_path
                    # Ensure this matches the attribute name in your memory_storage_configs items
                ))

        resources = {
            'requests': {},
            'limits': {}
        }

        if cpu_requests and cpu_limits:
            resources['requests']['cpu'] = K8SUnits(cpu_requests).as_str
            resources['limits']['cpu'] = K8SUnits(cpu_limits).as_str
        if memory_requests and memory_limits:
            resources['requests']['memory'] = K8SUnits(memory_requests).as_str
            resources['limits']['memory'] = K8SUnits(memory_limits).as_str
        if gpu_requests and gpu_limits:
            resources['requests']['nvidia.com/gpu'] = K8SUnits(gpu_limits).as_str
            resources['limits']['nvidia.com/gpu'] = K8SUnits(gpu_requests).as_str

        if security_context_config and security_context_config.enable_security_context:
            security_context = {
                "capabilities": {
                    "add": security_context_config.add_capabilities
                }
            }

        return client.V1Container(
            name=container_name,
            image=image_name,
            ports=[client.V1ContainerPort(container_port=port) for port in ports] if ports else [],
            env=env_vars,
            volume_mounts=volume_mounts,
            resources=client.V1ResourceRequirements(requests=resources['requests'], limits=resources['limits']),
            security_context=security_context
        )

    @staticmethod
    def create_container_ports(ports):
        # Check if self.ports is a single integer and convert it to a list if so
        ports = [ports] if isinstance(ports, int) else ports
        return [client.V1ContainerPort(container_port=port) for port in ports]

    @staticmethod
    def create_environment_variables(**envs):
        env_vars = []
        if envs:
            for env_var in envs:
                if isinstance(env_var, dict) and 'name' in env_var and 'value' in env_var:
                    # Ensure value is a string, convert if necessary
                    value = str(env_var['value']) if not isinstance(env_var['value'], str) else env_var['value']
                    env_vars.append(client.V1EnvVar(name=env_var['name'], value=value))
                elif isinstance(env_var, str):
                    # If env_var is a string, assume it's in "name=value" format
                    parts = env_var.split('=', 1)
                    if len(parts) == 2:
                        env_vars.append(client.V1EnvVar(name=parts[0], value=parts[1]))
                    else:
                        # For a plain string without '=', assign a generic name
                        env_vars.append(client.V1EnvVar(name=f"ENV_{env_var}", value=env_var))
                elif isinstance(env_var, (int, float)):
                    # For numeric types, convert to string and assign a generic name
                    env_vars.append(client.V1EnvVar(name=f"NUM_ENV_{env_var}", value=str(env_var)))
                else:
                    raise TypeError(f"Unsupported environment variable type: {type(env_var)}")
        return env_vars

    def create_pod_template(self, image_name, labels=None, deployment_name=None, project_name=None,
                            ports=None, service_account_name=None, pvc_mounts=None,
                            cpu_requests=None, cpu_limits=None, memory_requests=None,
                            memory_limits=None, gpu_requests=None, gpu_limits=None, memory_storage_configs=None,
                            security_context_config=None, entrypoint_args=None, entrypoint_envs=None, ):
        if labels is None:
            labels = {}
        if project_name:
            labels['project'] = project_name

        volumes = []

        if memory_storage_configs:
            for mem_storage in memory_storage_configs:
                # Initialize the memory_volume_spec without size_limit
                memory_volume_spec = client.V1EmptyDirVolumeSource(medium="Memory")

                # Conditionally set size_limit if specified
                if mem_storage.size_gb is not None:
                    # Attempt to set size_limit directly, converting explicitly to a string
                    memory_volume_spec.size_limit = mem_storage.size_gb.as_str  # Explicit string conversion

                volumes.append(client.V1Volume(
                    name=mem_storage.name,
                    empty_dir=memory_volume_spec  # The correct parameter name is empty_dir, not emptyDir
                ))

        if pvc_mounts:
            for mount in pvc_mounts:
                volumes.append(client.V1Volume(
                    name=mount['pvc_name'],  # Volume name matches the PVC name
                    persistent_volume_claim=client.V1PersistentVolumeClaimVolumeSource(claim_name=mount['pvc_name'])
                ))

        # Ensure image_name is not included in entrypoint_args or envs
        container = self.create_container(
            image_name=image_name,  # Ensure this is the only place where image_name is provided
            deployment_name=deployment_name,
            project_name=project_name,
            ports=ports,
            pvc_mounts=pvc_mounts,
            cpu_requests=cpu_requests,
            cpu_limits=cpu_limits,
            memory_requests=memory_requests,
            memory_limits=memory_limits,
            memory_storage_configs=memory_storage_configs,
            gpu_requests=gpu_requests,
            gpu_limits=gpu_limits,
            security_context_config=security_context_config,
            entrypoint_args=entrypoint_args,
            entrypoint_envs=entrypoint_envs
        )
        # Defining volumes for the pod spec based on PVC mounts

        pod_spec = client.V1PodSpec(
            containers=[container],
            service_account_name=service_account_name,
            volumes=volumes  # Including PVC volumes here
        )

        return client.V1PodTemplateSpec(
            metadata=client.V1ObjectMeta(labels=labels),
            spec=pod_spec
        )

    def create_pvc(self, pvc_name, pvc_size, pvc_access_mode, namespace):
        logger.info(f"Attempting to create PVC: {pvc_name} in namespace: {namespace}")
        pvc_manifest = {
            "apiVersion": "v1",
            "kind": "PersistentVolumeClaim",
            "metadata": {"name": pvc_name},
            "spec": {
                "accessModes": [pvc_access_mode],
                "resources": {"requests": {"storage": pvc_size}}
            }
        }
        self.core_v1_api.create_namespaced_persistent_volume_claim(namespace=namespace, body=pvc_manifest)
        logger.info(f"Created PVC '{pvc_name}' in namespace '{namespace}'.")

    def create_deployment_spec(self, image_name, labels=None, deployment_name=None, project_name=None, replicas=None,
                               ports=None, service_account_name=None, storage_configs=None,
                               cpu_requests=None, cpu_limits=None, memory_requests=None,
                               memory_limits=None, gpu_requests=None, gpu_limits=None, memory_storage_configs=None,
                               security_context_config=None, entrypoint_args=None,
                               entrypoint_envs=None):
        # Ensure pvc_mounts are prepared correctly from storage_configs if needed
        pvc_mounts = [{
            'pvc_name': sc.pvc_name,
            'mount_path': sc.pvc_mount_path
        } for sc in storage_configs if sc.create_pvc] if storage_configs else []

        # Create the pod template with correct arguments
        pod_template = self.create_pod_template(
            image_name=image_name,
            labels=labels,
            deployment_name=deployment_name,
            project_name=project_name,
            ports=ports,
            service_account_name=service_account_name,  # Use it here
            pvc_mounts=pvc_mounts,  # Assuming pvc_mounts is prepared earlier in the method
            cpu_requests=cpu_requests,
            cpu_limits=cpu_limits,
            memory_storage_configs=memory_storage_configs,
            memory_requests=memory_requests,
            memory_limits=memory_limits,
            gpu_requests=gpu_requests,
            gpu_limits=gpu_limits,
            security_context_config=security_context_config,
            entrypoint_args=entrypoint_args,
            entrypoint_envs=entrypoint_envs
        )

        # Create and return the deployment spec
        return client.V1DeploymentSpec(
            replicas=int(replicas),  # Ensure replicas is an int
            template=pod_template,
            selector={'matchLabels': pod_template.metadata.labels}
        )

    def create_deployment(self, image_name, labels=None, deployment_name=None, namespace=None, project_name=None,
                          replicas=None, ports=None, service_account_name=None, storage_configs=None,
                          cpu_requests=None, cpu_limits=None, memory_requests=None, memory_storage_configs=None,
                          memory_limits=None, gpu_requests=None, gpu_limits=None,
                          security_context_config=None, entrypoint_args=None, entrypoint_envs=None):
        if namespace is None:
            namespace = self.namespace

        if project_name is None:
            project_name = self.project_name

        # Generate a unique name for the deployment if it's not provided
        if deployment_name is None:
            deployment_name = self.generate_unique_deployment_name(base_name=image_name.split(':')[0],
                                                                   namespace=namespace)
            # Include the 'app' label set to the unique deployment name
            if labels is None:
                labels = {}
            labels['app'] = deployment_name  # Set the 'app' label to the unique deployment name

        deployment_name = self.generate_unique_deployment_name(deployment_name, namespace)

        deployment_spec = self.create_deployment_spec(
            image_name, labels=labels, deployment_name=deployment_name,
            project_name=project_name, replicas=replicas, ports=ports,
            service_account_name=service_account_name,  # Pass this
            storage_configs=storage_configs, cpu_requests=cpu_requests, cpu_limits=cpu_limits,
            memory_requests=memory_requests, memory_limits=memory_limits,
            memory_storage_configs=memory_storage_configs,
            gpu_requests=gpu_requests, gpu_limits=gpu_limits, security_context_config=security_context_config,
            entrypoint_args=entrypoint_args, entrypoint_envs=entrypoint_envs,
        )

        # Optionally add the project name to the deployment's metadata
        deployment_metadata = client.V1ObjectMeta(name=deployment_name, namespace=namespace,
                                                  labels={"project": project_name})

        # logger.info(f"Deployment {deployment_name} created in namespace {namespace}.")
        deployment = client.V1Deployment(
            api_version="apps/v1",
            kind="Deployment",
            metadata=deployment_metadata,
            spec=deployment_spec
        )
        logger.info(f"Created deployment object type: {type(deployment)}")
        return deployment

    def generate_unique_deployment_name(self, base_name, namespace):
        unique_name = base_name
        suffix = 1
        while True:
            try:
                self.apps_v1_api.read_namespaced_deployment(name=unique_name, namespace=namespace)
                # If the deployment exists, append/increment the suffix and try again
                unique_name = f"{base_name}-{suffix}"
                suffix += 1
            except ApiException as e:
                if e.status == 404:  # Not Found, the name is unique
                    return unique_name
                raise  # Reraise exceptions that are not related to the deployment not existing

    def apply_deployment(self, deployment, namespace=None):
        # logger.info(f"Applying deployment object type: {type(deployment)}")
        # Determine the namespace
        if namespace is None:
            namespace = self.namespace

        try:
            # Apply the deployment
            self.apps_v1_api.create_namespaced_deployment(body=deployment, namespace=namespace)
            logger.info(f"Successfully applied deployment in namespace '{namespace}'")

            # Construct the selector based on the deployment object type
            if isinstance(deployment, dict):
                # If deployment is a dictionary, use dictionary access
                selector = ','.join([f'{k}={v}' for k, v in deployment['spec']['selector']['matchLabels'].items()])
            else:
                # If deployment is an object, use attribute access
                selector = ','.join([f'{k}={v}' for k, v in deployment.spec.selector['matchLabels'].items()])

            w = watch.Watch()
            timeout = 300  # Define a timeout for the watch (e.g., 300 seconds)
            start = time.time()

            for event in w.stream(self.core_v1_api.list_namespaced_pod, namespace=namespace, label_selector=selector,
                                  timeout_seconds=60):
                pod = event['object']
                if pod.status.phase == 'Running':
                    # Extract and return pod info
                    pod_info = self.extract_pod_info(pod)
                    logger.info(f"Pod info: '{pod_info}'")
                    w.stop()
                    return pod_info

                if time.time() - start > timeout:
                    logger.error("Timeout waiting for pods to become ready.")
                    w.stop()
                    return None

        except ApiException as e:
            logger.exception(f"Exception when applying the deployment: {e}")
            return None

    @staticmethod
    def extract_pod_info(pod):
        # Extract pod name, namespace, ports, and PVCs
        pod_name = pod.metadata.name
        namespace = pod.metadata.namespace
        ports = [container.ports for container in pod.spec.containers]
        pvcs = [volume.persistent_volume_claim.claim_name for volume
                in pod.spec.volumes if volume.persistent_volume_claim]

        # Create a dynamic data class for PodInfo
        podinfo = make_dataclass('PodInfo', [('name', str), ('namespace', str), ('ports', list), ('pvcs', list)])
        return podinfo(name=pod_name, namespace=namespace, ports=ports, pvcs=pvcs)

        # TODO: return pod info - extract pod name and namespace from the applied_deployment object,
        # if needed, beam_k8s will query the server to retrieve all the necessary information

    def create_role_bindings(self, user_idm_configs):
        for config in user_idm_configs:
            if config.create_role_binding:
                role_binding = {
                    "apiVersion": "rbac.authorization.k8s.io/v1",
                    "kind": "RoleBinding",
                    "metadata": {
                        "name": config.role_binding_name,
                        "namespace": config.project_name  # Namespace is derived from project_name
                    },
                    "subjects": [{
                        "kind": "User",
                        "name": config.user_name,
                        "apiGroup": "rbac.authorization.k8s.io"
                    }],
                    "roleRef": {
                        "kind": "ClusterRole",
                        "name": config.role_name,
                        # Assuming 'admin' or equivalent ClusterRole that provides namespace-level admin access
                        "apiGroup": "rbac.authorization.k8s.io"
                    }
                }

                try:
                    self.dyn_client.resources.get(api_version='rbac.authorization.k8s.io/v1',
                                                  kind='RoleBinding').create(body=role_binding,
                                                                             namespace=config.project_name)
                    logger.info(
                        f"Admin role binding '{config.role_binding_name}' for user "
                        f"'{config.user_name}' created in namespace '{config.project_name}'.")
                except ApiException as e:
                    if e.status == 409:  # Conflict error - RoleBinding already exists
                        logger.info(
                            f"Role binding '{config.role_binding_name}' "
                            f"already exists in namespace '{config.project_name}', skipping.")
                    else:
                        logger.error(
                            f"Failed to create admin role binding for '{config.user_name}' "
                            f"in namespace '{config.project_name}': {e}")

    def create_service(self, base_name, namespace, ports, labels, service_type):
        # Initialize the service name with the base name
        service_name = base_name

        # Ensure ports is a list, even if it's None or empty
        if ports is None:
            ports = []

        # Check if the service already exists
        try:
            existing_service = self.core_v1_api.read_namespaced_service(name=base_name, namespace=namespace)
            if existing_service:
                print(f"Service '{base_name}' already exists in namespace '{namespace}'. Generating a unique name.")
                # Generate a unique name for the service
                service_name = self.generate_unique_service_name(base_name, namespace)
        except client.exceptions.ApiException as e:
            if e.status != 404:  # If the error is not 'Not Found', raise it
                raise

        # Do not override 'app' label if it's already set in the labels dictionary
        if 'app' not in labels:
            labels['app'] = service_name

        # Define service metadata with the unique name
        metadata = client.V1ObjectMeta(name=service_name, labels=labels)

        # Dynamically create service ports from the ports list, including unique names for each
        service_ports = []
        for idx, port in enumerate(ports):
            port_name = f"{service_name}-port-{idx}-{port}"
            service_ports.append(client.V1ServicePort(name=port_name, port=port, target_port=port))

        # Define service spec with dynamically set ports
        service_spec = client.V1ServiceSpec(
            ports=service_ports,
            selector=labels,
            type=service_type
        )

        # Create the Service object with the unique name
        service = client.V1Service(
            api_version="v1",
            kind="Service",
            metadata=metadata,
            spec=service_spec
        )

        # Create the service in the specified namespace
        try:
            self.core_v1_api.create_namespaced_service(namespace=namespace, body=service)
            print(f"Service '{service_name}' created successfully in namespace '{namespace}'.")
            logger.info(
                f"Service '{service_name}' of type '{service_type}' created with ports: "
                f"{', '.join([f'Port: {port.port}, TargetPort: {port.target_port}' for port in service_ports])}")
        except client.exceptions.ApiException as e:
            print(f"Failed to create service '{service_name}' in namespace '{namespace}': {e}")

    def generate_unique_service_name(self, base_name, namespace):
        unique_name = base_name
        suffix = 1
        while True:
            try:
                self.core_v1_api.read_namespaced_service(name=unique_name, namespace=namespace)
                # If the service exists, append/increment the suffix and try again
                unique_name = f"{base_name}-{suffix}"
                suffix += 1
            except client.exceptions.ApiException as e:
                if e.status == 404:  # Not Found, the name is unique
                    return unique_name
                raise  # Reraise exceptions that are not related to the service not existing

    def delete_service(self, deployment_name, namespace=None):
        from kubernetes.client import V1DeleteOptions

        if namespace is None:
            namespace = self.namespace

        try:
            # Get the service associated with the deployment
            service_list = self.core_v1_api.list_namespaced_service(namespace=namespace)
            for service in service_list.items:
                if service.metadata.labels.get("app") == deployment_name:
                    service_name = service.metadata.name
                    # Use the core_v1_api to delete the Service
                    self.core_v1_api.delete_namespaced_service(
                        name=service_name,
                        namespace=namespace,
                        body=V1DeleteOptions()
                    )
                    logger.info(f"Deleted service '{service_name}' from namespace '{namespace}'.")
                    return  # Exit the loop once the service is deleted
        except ApiException as e:
            logger.error(f"Error deleting service for deployment '{deployment_name}': {e}")

    def delete_services_by_label_selector(self, label_selector, namespace=None):
        if namespace is None:
            namespace = self.namespace

        try:
            services = self.core_v1_api.list_namespaced_service(namespace=namespace, label_selector=label_selector)
            for service in services.items:
                service_name = service.metadata.name
                self.core_v1_api.delete_namespaced_service(
                    name=service_name,
                    namespace=namespace,
                    body=V1DeleteOptions()
                )
                logger.info(f"Deleted service '{service_name}' in namespace '{namespace}'")
        except Exception as e:
            logger.error(
                f"Error deleting services with label selector '{label_selector}' in namespace '{namespace}': {e}")

    def create_route(self, service_name, namespace, protocol, port):
        from openshift.dynamic.exceptions import NotFoundError
        from openshift.dynamic import DynamicClient

        dyn_client = DynamicClient(self.api_client)

        # Get the Route resource from the OpenShift API
        route_resource = dyn_client.resources.get(api_version='route.openshift.io/v1', kind='Route')

        try:
            # Try to get the existing route
            existing_route = route_resource.get(name=service_name, namespace=namespace)
            # If the route exists, log a message and return
            logger.info(f"Route {service_name} already exists in namespace {namespace}, skipping creation.")
            return
        except NotFoundError:
            # The route does not exist, proceed with creation
            logger.info(f"Route {service_name} does not exist in namespace {namespace}, proceeding with creation.")
        except Exception as e:
            # Handle other exceptions that are not related to route not found
            logger.error(f"Error checking route {service_name} in namespace {namespace}: {e}")
            return

        # Define the route manifest for creation
        route_manifest = {
            "apiVersion": "route.openshift.io/v1",
            "kind": "Route",
            "metadata": {
                "name": service_name,
                "namespace": namespace
            },
            "spec": {
                "to": {
                    "kind": "Service",
                    "name": service_name
                },
                "port": {
                    "targetPort": port  # Use numeric port
                }
            }
        }

        # Add TLS termination if protocol is 'https'
        if protocol.lower() == 'https':
            route_manifest["spec"]["tls"] = {
                "termination": "edge"
            }

        # Attempt to create the route
        try:
            created_route = route_resource.create(body=route_manifest, namespace=namespace)
            logger.info(f"Route for service {service_name} created successfully in namespace {namespace}.")
            # Print the DNS name of the route
            dns_name = created_route.spec.host  # Accessing the DNS name from the route response
            logger.info(f"The DNS name of the created route is: {dns_name}")  # Log the DNS name
        except Exception as e:
            logger.error(f"Failed to create route for service {service_name} in namespace {namespace}: {e}")

    def delete_route(self, route_name, namespace):
        from openshift.dynamic.exceptions import NotFoundError
        from openshift.dynamic import DynamicClient

        dyn_client = DynamicClient(self.api_client)

        # Get the Route resource from the OpenShift API
        route_resource = dyn_client.resources.get(api_version='route.openshift.io/v1', kind='Route')

        try:
            # Try to get the existing route
            existing_route = route_resource.get(name=route_name, namespace=namespace)
            # If the route exists, delete it
            existing_route.delete()
            logger.info(f"Deleted route '{route_name}' from namespace '{namespace}'.")
        except NotFoundError:
            # If the route doesn't exist, log a message
            logger.info(f"Route '{route_name}' does not exist in namespace '{namespace}'.")
        except Exception as e:
            # Handle other exceptions
            logger.error(f"Error deleting route '{route_name}' in namespace '{namespace}': {e}")

    def create_ingress(self, service_configs, default_host=None, default_path="/", default_tls_secret=None):
        from kubernetes.client import (V1Ingress, V1IngressSpec, V1IngressRule, V1HTTPIngressRuleValue,
                                       V1HTTPIngressPath,
                                       V1IngressBackend, V1ServiceBackendPort, V1IngressTLS, V1ObjectMeta)
        # Initialize the NetworkingV1Api
        networking_v1_api = client.NetworkingV1Api()

        for svc_config in service_configs:
            if not svc_config.create_ingress:
                continue  # Skip if create_ingress is False for this service config

            # Use specific values from svc_config or fall back to default parameters
            host = svc_config.ingress_host if svc_config.ingress_host else f"{svc_config.service_name}.example.com"
            path = svc_config.ingress_path if svc_config.ingress_path else default_path
            tls_secret = svc_config.ingress_tls_secret if svc_config.ingress_tls_secret else default_tls_secret

            # Define Ingress metadata
            metadata = V1ObjectMeta(
                name=f"{svc_config.service_name}-ingress",
                namespace=self.namespace,
                annotations={
                    "kubernetes.io/ingress.class": "nginx",
                    "nginx.ingress.kubernetes.io/ssl-redirect": "true",
                    "nginx.ingress.kubernetes.io/force-ssl-redirect": "true",
                }
            )

            # Define the backend service
            backend = V1IngressBackend(
                service=V1ServiceBackendPort(
                    name=svc_config.port_name,
                    number=svc_config.port
                )
            )

            # Define the Ingress rule
            rule = V1IngressRule(
                host=host,
                http=V1HTTPIngressRuleValue(
                    paths=[
                        V1HTTPIngressPath(
                            path=path,
                            path_type="Prefix",
                            backend=backend
                        )
                    ]
                )
            )

            # Define Ingress Spec with optional TLS configuration
            spec = V1IngressSpec(rules=[rule])
            if tls_secret:
                spec.tls = [
                    V1IngressTLS(
                        hosts=[host],
                        secret_name=tls_secret
                    )
                ]

            # Create the Ingress object
            ingress = V1Ingress(
                api_version="networking.k8s.io/v1",
                kind="Ingress",
                metadata=metadata,
                spec=spec
            )

            # Use the NetworkingV1Api to create the Ingress
            try:
                networking_v1_api.create_namespaced_ingress(namespace=self.namespace, body=ingress)
                logger.info(
                    f"Ingress for service {svc_config.service_name} "
                    f"created successfully in namespace {self.namespace}.")
            except Exception as e:
                logger.error(
                    f"Failed to create Ingress for service {svc_config.service_name} "
                    f"in namespace {self.namespace}: {e}")

    def delete_ingress(self, service_name):
        from kubernetes.client import NetworkingV1Api, V1DeleteOptions

        try:
            # Initialize the NetworkingV1Api
            networking_v1_api = NetworkingV1Api(self.api_client)

            # Use the NetworkingV1Api to delete the Ingress
            networking_v1_api.delete_namespaced_ingress(
                name=f"{service_name}-ingress",
                namespace=self.namespace,
                body=V1DeleteOptions()
            )
            logger.info(f"Ingress for service {service_name} deleted successfully.")
        except Exception as e:
            logger.error(f"Failed to delete Ingress for service {service_name}: {e}")

    def print_pod_node_info(self, deployment_name):
        # Step 1: Find the deployment to get its selector
        deployment = self.apps_v1_api.read_namespaced_deployment(deployment_name, self.namespace)
        selector = ','.join([f'{k}={v}' for k, v in deployment.spec.selector.match_labels.items()])

        # Step 2: List all pods in the deployment using the selector
        pods = self.core_v1_api.list_namespaced_pod(namespace=self.namespace, label_selector=selector)

        # Collect unique node names where the pods are scheduled
        node_names = set()
        for pod in pods.items:
            node_name = pod.spec.node_name
            node_names.add(node_name)
            logger.info(f"Pod: {pod.metadata.name} is running on Node: {node_name}")

        # Step 3: List services and check if they target the deployment
        services = self.core_v1_api.list_namespaced_service(namespace=self.namespace)
        for service in services.items:
            if service.spec.type == "NodePort":
                for port in service.spec.ports:
                    if port.node_port:
                        logger.info(
                            f"NodePort Service: {service.metadata.name}, Port: {port.port}, NodePort: {port.node_port}")
                        # Since we can't get the external IP, we'll just inform about using the cluster's node IPs
                        logger.info(
                            "To access the NodePort service, use the IP "
                            "address of any cluster node with the listed NodePort.")

        # Inform about the limitation regarding node IPs
        logger.info(
            "Node external IPs cannot be retrieved with namespace-scoped "
            "permissions. Use known node IPs to access NodePort services.")

    def get_internal_endpoints_with_nodeport(self, namespace):
        endpoints = []
        try:
            services = self.core_v1_api.list_namespaced_service(namespace=namespace)
            nodes = self.core_v1_api.list_node()
            node_ips = {node.metadata.name:
                            [address.address for address in node.status.addresses if address.type == "InternalIP"][0]
                        for node in nodes.items}

            for service in services.items:
                if service.spec.type == "NodePort":
                    for port in service.spec.ports:
                        for node_name, node_ip in node_ips.items():
                            endpoint = {'node_ip': node_ip, 'node_port': port.node_port,
                                        'service_name': service.metadata.name}
                            if endpoint not in endpoints:  # Check for uniqueness
                                endpoints.append(endpoint)
                                print(
                                    f"Debug: Adding endpoint for service {service.metadata.name} "
                                    f"on node {node_name} - {endpoint}")

        except client.exceptions.ApiException as e:
            print(f"Failed to retrieve services or nodes in namespace '{namespace}': {e}")

        return endpoints

    def query_available_resources(self):
        total_resources = {'cpu': '0', 'memory': '0', 'nvidia.com/gpu': '0', 'amd.com/gpu': '0', 'storage': '0Gi'}
        node_list = self.core_v1_api.list_node()

        # Summing up the allocatable CPU, memory, and GPU resources from each node
        for node in node_list.items:
            for key, quantity in node.status.allocatable.items():
                if key in ['cpu', 'memory', 'nvidia.com/gpu', 'amd.com/gpu']:
                    if quantity.endswith('m'):  # Handle milliCPU
                        total_resources[key] = str(
                            int(total_resources.get(key, '0')) + int(float(quantity.rstrip('m')) / 1000))
                    else:
                        total_resources[key] = str(
                            int(total_resources.get(key, '0')) + int(quantity.strip('Ki')))

        # Summing up the storage requests for all PVCs in the namespace
        pvc_list = self.core_v1_api.list_namespaced_persistent_volume_claim(namespace=self.namespace)
        for pvc in pvc_list.items:
            for key, quantity in pvc.spec.resources.requests.items():
                if key == 'storage':
                    total_resources['storage'] = str(
                        int(total_resources['storage'].strip('Gi')) + int(quantity.strip('Gi'))) + 'Gi'

        # Remove resources with a count of '0'
        total_resources = {key: value for key, value in total_resources.items() if value != '0'}

        logger.info(f"Total Available Resources in the Namespace '{self.namespace}': {total_resources}")
        return total_resources

    def execute_command_in_pod(self, namespace, pod_name, command):
        """
        Execute a command in a pod.

        :param namespace: The namespace of the pod.
        :param pod_name: The name of the pod where the command will be executed.
        :param command: The command to execute inside the pod, as a list of strings.
        :return: The output from the command execution.
        """
        # Ensure that a list is provided for the command
        if not isinstance(command, list):
            command = [command]

        # Executing the command
        response = stream(self.core_v1_api.connect_get_namespaced_pod_exec,
                          pod_name,
                          namespace,
                          command=command,
                          stderr=True,
                          stdin=False,
                          stdout=True,
                          tty=False)
        return response

    def get_pod_info(self, pod_name, namespace):
        """Retrieve information about a specific pod."""
        try:
            return self.core_v1_api.read_namespaced_pod(name=pod_name, namespace=namespace or self.namespace)
        except ApiException as e:
            logger.error(f"Failed to get pod info for {pod_name} in {namespace}: {e}")
            return None

    def get_pod_logs(self, pod_name, namespace=None, **kwargs):
        """Retrieve logs for a specific pod."""
        try:
            return self.core_v1_api.read_namespaced_pod_log(name=pod_name, namespace=namespace or self.namespace,
                                                            **kwargs)
        except ApiException as e:
            logger.error(f"Failed to get logs for {pod_name} in {namespace}: {e}")
            return None

    def get_pod_resources(self, pod_name, namespace=None):
        """Get resource usage for a specific pod using the dynamic client."""
        try:
            resource = self.dyn_client.resources.get(api_version='metrics.k8s.io/v1beta1', kind='PodMetrics')
            return resource.get(name=pod_name, namespace=namespace or self.namespace)
        except ApiException as e:
            logger.error(f"Failed to get resources for {pod_name} in {namespace}: {e}")
            return None

    def stop_pod(self, pod_name, namespace=None):
        """Stop a specific pod. This is usually done by deleting the pod."""
        try:
            self.core_v1_api.delete_namespaced_pod(name=pod_name, namespace=namespace or self.namespace,
                                                   body=V1DeleteOptions())
            logger.info(f"Pod {pod_name} in {namespace} deleted successfully.")
        except ApiException as e:
            logger.error(f"Failed to delete pod {pod_name} in {namespace}: {e}")

    def start_pod(self, pod_name, namespace=None, pod_body=None):
        """Start a specific pod by creating it. `pod_body` should be a V1Pod manifest."""
        try:
            self.core_v1_api.create_namespaced_pod(namespace=namespace or self.namespace, body=pod_body)
            logger.info(f"Pod {pod_name} in {namespace} started successfully.")
        except ApiException as e:
            logger.error(f"Failed to create pod {pod_name} in {namespace}: {e}")

    # def list_pods(self):
    #     label_selector = f"app={self.deployment_name}"
    #     pods = self.core_v1_api.list_namespaced_pod(namespace=self.namespace, label_selector=label_selector)
    #     for pod in pods.items:
    #         print(f"Pod Name: {pod.metadata.name}, Pod Status: {pod.status.phase}")




import torch
from .core import BeamSimilarity, Similarities
from ..utils import check_type, as_tensor, beam_device, tqdm_beam as tqdm, lazy_property


class SparseSimilarity(BeamSimilarity):
    """
    The `SparseSimilarity` class is a processor that computes similarity between sparse vectors.

    Args:
        metric (str): The similarity metric to use. Possible values are 'cosine', 'prod', 'l2', and 'max'.
                      Default is 'cosine'.
        layout (str): The layout format of the sparse vectors. Possible values are 'coo' and 'csr'. Default is 'coo'.
        vec_size (int): The size of the vectors. Required if the layout is 'csr', otherwise optional.
        device (str): The device to use for computation. Default is None, which means using the default device.
        k (int): The number of nearest neighbors to search for. Default is 1.
        q (float): The quantile value to use for the 'quantile' metric. Default is 0.9.

    Methods:
        reset()
            Reset the state of the processor.

        sparse_tensor(r, c, v)
            Convert coordinate, row, column, and value data into a sparse tensor.

            Args:
                r (Tensor): The row indices.
                c (Tensor): The column indices.
                v (Tensor): The values.

            Returns:
                SparseTensor: The sparse tensor.

        index
            Get the current index tensor.

        scipy_to_row_col_val(x)
            Convert a sparse matrix in the scipy sparse format to row, column, and value data.

            Args:
                x (scipy.sparse.spmatrix): The sparse matrix.

            Returns:
                Tensor: The row indices.
                Tensor: The column indices.
                Tensor: The values.

        to_sparse(x)
            Convert input data to a sparse tensor.

            Args:
                x (Tensor, numpy.ndarray, scipy.sparse.spmatrix, dict, tuple): The input data.

            Returns:
                SparseTensor: The sparse tensor.

        add(x)
            Add a sparse vector to the index.

            Args:
                x (Tensor, numpy.ndarray, scipy.sparse.spmatrix, dict, tuple): The input sparse vector.

        search(x, k=None)
            Search for the nearest neighbors of a sparse vector.

            Args:
                x (SparseTensor, Tensor, numpy.ndarray, scipy.sparse.spmatrix, dict, tuple): The query sparse vector.
                k (int): The number of nearest neighbors to search for. If not specified, use the default value.

            Returns:
                Tensor: The distances to the nearest neighbors.
                Tensor: The indices of the nearest neighbors.
    """
    def __init__(self, *args, metric='cosine', layout='coo', vec_size=None, device=None, quantile=.9, **kwargs):

        super().__init__(*args, metric=metric, layout=layout, vec_size=vec_size, device=device, quantile=quantile,
                         **kwargs)
        # possible similarity metrics: cosine, prod, l2, max
        self.metric = self.get_hparam('metric', metric)
        self.layout = self.get_hparam('layout', layout)
        self.device = beam_device(self.get_hparam('device', device))
        self.vec_size = vec_size
        self.index = None
        self.vectors = None
        self.quantile = self.get_hparam('quantile', quantile)

    def reset(self):
        self.index = None
        self.vectors = None

    def sparse_tensor(self, r, c, v,):
        device = self.device
        size = (r.max() + 1, self.vec_size)

        r, c, v = as_tensor([r, c, v], device=device)

        if self.layout == 'coo':
            return torch.sparse_coo_tensor(torch.stack([r, c]), v, size=size, device=device)

        if self.layout == 'csr':
            return torch.sparse_csr_tensor(r, c, v, size=size, device=device)

        raise ValueError(f"Unknown format: {self.layout}")

    @staticmethod
    def scipy_to_row_col_val(x):

        r, c = x.nonzero()
        return r, c, x.data

    def to_sparse(self, x):

        x_type = check_type(x)

        if x_type.minor == 'scipy_sparse':
            r, c, v = self.scipy_to_row_col_val(x)
            x = self.sparse_tensor(r, c, v)

        elif x_type.minor in ['tensor', 'numpy']:

            if x_type.minor == 'numpy':
                x = as_tensor(x)

            if self.layout == 'coo':
                x = x.to_sparse_coo()
            elif self.layout == 'csr':
                x = x.to_sparse_csr()
            else:
                raise ValueError(f"Unknown format: {self.layout}")

        elif x_type.minor == 'dict':
            x = self.sparse_tensor(x['row'], x['col'], x['val'])

        elif x_type.minor == 'tuple':
            x = self.sparse_tensor(x[0], x[1], x[2])

        else:
            raise ValueError(f"Unsupported type: {x_type}")

        return x

    def add(self, x, index=None, **kwargs):

        x = self.to_sparse(x)
        if self.vectors is None:
            self.vectors = x
        else:
            self.vectors = torch.cat([self.vectors, x])

        if index is not None:
            if self.index is None:
                self.index = index
            else:
                self.index = torch.cat([self.index, index])
        else:
            if self.index is None:
                self.index = torch.arange(len(x), device=self.device)
            else:
                index = torch.arange(len(x), device=self.device) + self.index.max() + 1
                self.index = torch.cat([self.index, index])

    def train(self, x=None):
        raise NotImplementedError

    def search(self, x, k=1, **kwargs):

        x = self.to_sparse(x)

        if self.metric in ['cosine', 'l2', 'prod']:

            if self.layout == 'csr':
                x = x.to_dense()

            ab = self.index @ x.T

            if self.metric in ['l2', 'cosine']:

                a2 = (self.index * self.index).sum(dim=1, keepdim=True)
                b2 = (x * x).sum(dim=1, keepdim=True)

                if self.metric == 'cosine':

                    s = 1 / torch.sqrt(a2 @ b2.T).to_dense()
                    dist = - ab * s
                else:
                    dist = a2 + b2 - 2 * ab

            elif self.metric == 'prod':
                dist = -ab

            dist = dist.to_dense()

        elif self.metric in ['max', 'quantile']:
            x = x.to_dense()

            def metric(x):
                if self.metric == 'max':
                    return x.max()
                elif self.metric == 'quantile':
                    return x.quantile(self.quantile)
                else:
                    raise ValueError(f"Unknown metric: {self.metric}")

            dist = []
            for xi in x:
                d = self.index * xi.unsqueeze(0)
                i = d._indices()
                v = d._values()

                dist.append(as_tensor([metric(v[i[0] == j]) for j in range(len(self.index))]))

            dist = -torch.stack(dist, dim=1)

        topk = torch.topk(dist, k, dim=0, largest=False, sorted=True)
        I = topk.indices.T
        D = topk.values.T

        return Similarities(index=self.index[I], distance=D, metric=self.metric, model='sparse')



from .models import TGILLM, FastChatLLM, FastAPILLM, HuggingFaceLLM, FastAPIDPLLM
from .openai import OpenAILLM
from ..path import beam_key, BeamURL


def beam_llm(url, username=None, hostname=None, port=None, api_key=None, **kwargs):
    """
    @param url: The URL of the language model. Can be a string or an instance of BeamURL.
    @param username: The username to authenticate with the language model.
    @param hostname: The hostname of the language model.
    @param port: The port number of the language model.
    @param api_key: The API key to authenticate with the language model.
    @param kwargs: Additional keyword arguments to pass to the language model constructor.
    @return: An instance of the corresponding language model class based on the URL's protocol.

    """
    if type(url) != str:
        return url

    url = BeamURL.from_string(url)

    if url.hostname is not None:
        hostname = url.hostname

    if url.port is not None:
        port = url.port

    if url.username is not None:
        username = url.username

    query = url.query
    for k, v in query.items():
        kwargs[k] = v

    if api_key is None and 'api_key' in kwargs:
        api_key = kwargs.pop('api_key')

    model = url.path
    model = model.strip('/')
    if not model:
        model = None

    if url.protocol == 'openai':

        api_key = beam_key('OPENAI_API_KEY', api_key)
        return OpenAILLM(model=model, api_key=api_key, **kwargs)

    elif url.protocol == 'fastchat':
        return FastChatLLM(model=model, hostname=hostname, port=port, **kwargs)

    elif url.protocol == 'huggingface':
        return HuggingFaceLLM(model=model, **kwargs)

    elif url.protocol == 'fastapi':
        return FastAPILLM(model=model, hostname=hostname, port=port, username=username, **kwargs)

    elif url.protocol == 'fastapi-dp':
        return FastAPIDPLLM(model=model, hostname=hostname, port=port, username=username, **kwargs)

    elif url.protocol == 'tgi':
        return TGILLM(model=model, hostname=hostname, port=port, username=username, **kwargs)

    else:
        raise NotImplementedError


import json
import math
from json import JSONDecodeError
from typing import Optional, Any
import requests
import pandas as pd
import numpy as np

from .hf_conversation import Conversation
from ..logger import beam_logger as logger
from .core import BeamLLM, CompletionObject
from pydantic import BaseModel, Field, PrivateAttr
from ..path import beam_key, normalize_host

from .utils import get_conversation_template
from ..utils import lazy_property, strip_suffix
from .openai import OpenAIBase


class FCConversationLLM(BeamLLM):

    def get_prompt(self, messages):

        conv = get_conversation_template(self.adapter)
        for m in messages:

            if m['role'] == 'system':

                role = m['system_name'] if 'system_name' in m else 'system'
                content = m['content']

            else:

                if m['role'] == 'user':
                    role = conv.roles[0]
                else:
                    role = conv.roles[1]

                content = m['content']

            conv.append_message(role, content)

        conv.append_message(conv.roles[1], None)

        return conv.get_prompt()

    @property
    def is_chat(self):
        return True

    @property
    def is_completions(self):
        return True


class TGILLM(FCConversationLLM):

    _info: Any = PrivateAttr(default=None)
    _client: Any = PrivateAttr(default=None)
    timeout: Optional[float] = Field(None)

    def __init__(self, hostname=None, port=None, *args, adapter=None, timeout=None, **kwargs):

        api_base = f"http://{normalize_host(hostname, port)}"

        client_kwargs = {}
        if timeout is not None:
            client_kwargs['timeout'] = timeout

        from text_generation import Client
        self._client = Client(api_base, **client_kwargs)

        req = requests.get(f"{api_base}/info")
        self._info = json.loads(req.text)

        kwargs['model'] = self._info['model_id']
        kwargs['scheme'] = 'tgi'

        if adapter is None:
            adapter = 'raw'
        super().__init__(*args, adapter=adapter, **kwargs)

    def update_usage(self, response):

        self.usage["prompt_tokens"] += 0
        self.usage["completion_tokens"] += response.details.generated_tokens
        self.usage["total_tokens"] += 0 + response.details.generated_tokens

    def openai_format(self, res, **kwargs):
        return super().openai_format(res, tokens=res.response.details.tokens,
                                     completion_tokens=res.response.details.generated_tokens,
                                     total_tokens=res.response.details.generated_tokens)

    @property
    def is_chat(self):
        return False

    @property
    def is_completions(self):
        return True

    def process_kwargs(self, prompt, **kwargs):

        processed_kwargs = {}

        max_tokens = kwargs.pop('max_tokens', None)
        max_new_tokens = None

        if max_tokens is not None:
            max_new_tokens = max_tokens - self.len_function(prompt)

        max_new_tokens = kwargs.pop('max_new_tokens', max_new_tokens)
        if max_new_tokens is not None:
            processed_kwargs['max_new_tokens'] = max_new_tokens

        temperature = kwargs.pop('temperature', None)
        if temperature is not None:
            processed_kwargs['temperature'] = temperature

        top_p = kwargs.pop('top_p', None)
        if top_p is not None and 0 < top_p < 1:
            processed_kwargs['top_p'] = top_p

        best_of = kwargs.pop('n', None)
        if best_of is not None and best_of > 1:
            processed_kwargs['best_of'] = best_of

        stop_sequences = kwargs.pop('stop', None)
        if stop_sequences is None:
            stop_sequences = []
        elif type(stop_sequences) is str:
            stop_sequences = [stop_sequences]

        if self.stop_sequence is not None:
            stop_sequences.append(self.stop_sequence)

        if len(stop_sequences) > 0:
            processed_kwargs['stop_sequences'] = stop_sequences

        if 'top_k' in kwargs:
            processed_kwargs['top_k'] = kwargs.pop('top_k')

        if 'truncate' in kwargs:
            processed_kwargs['truncate'] = kwargs.pop('truncate')

        if 'typical_p' in kwargs:
            processed_kwargs['typical_p'] = kwargs.pop('typical_p')

        if 'watermark' in kwargs:
            processed_kwargs['watermark'] = kwargs.pop('watermark')

        if 'repetition_penalty' in kwargs:
            processed_kwargs['repetition_penalty'] = kwargs.pop('repetition_penalty')

        decoder_input_details = kwargs.pop('logprobs', None)
        if decoder_input_details is not None:
            processed_kwargs['decoder_input_details'] = decoder_input_details

        if 'seed' in kwargs:
            processed_kwargs['seed'] = kwargs.pop('seed')

        do_sample = None
        if temperature is not None and temperature > 0:
            do_sample = True
        do_sample = kwargs.pop('do_sample', do_sample)

        if do_sample is not None:
            processed_kwargs['do_sample'] = do_sample

        return processed_kwargs

    def _completion(self, prompt=None, **kwargs):

        prompt = self.get_prompt([{'role': 'user', 'content': prompt}])
        generate_kwargs = self.process_kwargs(prompt, **kwargs)
        res = self._client.generate(prompt, **generate_kwargs)
        return CompletionObject(prompt=prompt, kwargs=kwargs, response=res)

    def _chat_completion(self, messages=None, **kwargs):

        prompt = self.get_prompt(messages)
        generate_kwargs = self.process_kwargs(prompt, **kwargs)
        res = self._client.generate(prompt, **generate_kwargs)
        return CompletionObject(prompt=prompt, kwargs=kwargs, response=res)

    def extract_text(self, res):

        res = res.response
        text = res.generated_text
        text = strip_suffix(text, self.stop_sequence)
        return text


class FastChatLLM(OpenAIBase):

    def __init__(self, model=None, hostname=None, port=None, *args, **kwargs):

        api_base = f"http://{normalize_host(hostname, port)}/v1"
        api_key = "EMPTY"  # Not support yet
        organization = "EMPTY"  # Not support yet

        kwargs['scheme'] = 'fastchat'
        super().__init__(*args, api_key=api_key, api_base=api_base, organization=organization, model=model, **kwargs)

        self.model = model

    @property
    def is_chat(self):
        return True


class FastAPILLM(FCConversationLLM):

    normalized_hostname: Optional[str] = Field(None)
    headers: Optional[dict] = Field(None)
    consumer: Optional[str] = Field(None)
    protocol: Optional[str] = Field(None)
    _models: Any = PrivateAttr(default=None)

    def __init__(self, *args, model=None, hostname=None, port=None, username=None, protocol='https', **kwargs):

        kwargs['scheme'] = 'fastapi'
        super().__init__(*args, hostname=hostname, port=port, model=model, **kwargs)

        self.consumer = username
        self.normalized_hostname = normalize_host(hostname, port)
        self._models = None
        self.headers = {'Content-Type': 'application/json'}
        self.protocol = protocol

        from requests.packages.urllib3.exceptions import InsecureRequestWarning
        # Suppress only the single InsecureRequestWarning from urllib3
        requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

    @property
    def models(self):
        if self._models is None:
            res = requests.get(f"{self.protocol}://{self.normalized_hostname}/models", headers=self.headers, verify=False)
            self._models = res.json()
        return self._models

    @property
    def is_chat(self):
        return True

    @property
    def is_completions(self):
        return True

    def process_kwargs(self, prompt, **kwargs):

        kwargs_processed = {}

        max_tokens = kwargs.pop('max_tokens', None)
        max_new_tokens = None

        if max_tokens is not None:
            max_new_tokens = max_tokens - self.len_function(prompt)

        max_new_tokens = kwargs.pop('max_new_tokens', max_new_tokens)
        if max_new_tokens is not None:
            kwargs_processed['max_new_tokens'] = max_new_tokens

        temperature = kwargs.pop('temperature', None)
        if temperature is not None:
            kwargs_processed['temp'] = temperature

        if 'stop_criteria' in kwargs:
            kwargs_processed['stop_criteria'] = kwargs.pop('stop_criteria')

        kwargs_processed['stream'] = kwargs.pop('stream', False)

        return kwargs_processed

    def _chat_completion(self, messages=None, **kwargs):

        prompt = self.get_prompt(messages)
        return self._completion_internal(prompt, **kwargs)

    def _completion(self, prompt=None, **kwargs):

        prompt = self.get_prompt([{'role': 'user', 'content': prompt}])
        return self._completion_internal(prompt, **kwargs)

    def _stream_generator(self, d):

        with requests.post(f"{self.protocol}://{self.normalized_hostname}/predict/stream", headers=self.headers, json=d,
                           stream=True, verify=False) as response:
            for chunk in response.iter_content(chunk_size=1024):

                chunks = chunk.split("\0")
                for c in chunks:
                    if not len(c):
                        continue
                    try:
                        yield json.loads(c.decode())
                    except JSONDecodeError:
                        logger.error(f"Bad chunk: {c}")
                        yield c

    def _completion_internal(self, prompt, stream=False, **kwargs):

        d = dict(model_name=self.model, consumer=self.consumer, input=prompt,
                 hyper_params=self.process_kwargs(prompt, **kwargs))

        if stream:
            res = self._stream_generator(d)
            return CompletionObject(prompt=d['input'], kwargs=d, response=res)

        else:
            res = requests.post(f"{self.protocol}://{self.normalized_hostname}/predict/once", headers=self.headers, json=d,
                                verify=False)
            return CompletionObject(prompt=d['input'], kwargs=d, response=res.json())

    def verify_response(self, res):

        if res.stream:
            return True

        try:

            if not res.response['is_done']:
                logger.warning(f"Model {self.model} is_done=False.")

            assert 'res' in res.response, f"Response does not contain 'res' key"

        except Exception as e:
            logger.error(f"Error in response: {res.response}")
            raise e
        return True

    def extract_text(self, res):
        return res.response['res']


class FastAPIDPLLM(FastAPILLM):

    stop_token_ids: Optional[list] = Field(None)
    stop: Optional[str] = Field(None)
    application: Optional[str] = Field(None)
    task: Optional[str] = Field(None)
    network: Optional[str] = Field(None)
    cut_long_prompt: Optional[bool] = Field(None)
    echo: Optional[bool] = Field(None)
    worker_generate_stream_ep: Optional[str] = Field(None)
    streaming_delimiter: Optional[bytes] = Field(None)
    retrials: Optional[int] = Field(None)
    timeout: Optional[int] = Field(None)

    def __init__(self, *args, application='Botson', task='None', network='NH', cut_long_prompt=False, echo=False,
                 worker_generate_stream_ep='worker_generate_stream/', streaming_delimiter=b"\0",
                 retrials=None, stop_token_ids=None, protocol='http', stop=None, timeout=60, **kwargs):

        kwargs['scheme'] = 'fastapi-dp'
        super().__init__(*args, protocol=protocol, **kwargs)

        self.application = application
        self.task = task
        self.network = network
        self.cut_long_prompt = cut_long_prompt
        self.echo = echo
        self.worker_generate_stream_ep = worker_generate_stream_ep
        self.streaming_delimiter = streaming_delimiter
        self.stop_token_ids = stop_token_ids
        self.stop = stop
        self.retrials = retrials
        self.timeout = timeout
        self.headers = {'Content-Type': 'application/json', 'accept': 'application/json'}

    def process_kwargs(self, prompt, number_of_generations=None, **kwargs):
        kwargs_processed = super().process_kwargs(prompt, **kwargs)
        max_new_tokens = kwargs_processed['max_new_tokens']
        kwargs_processed['number_of_generations'] = int(math.ceil(max_new_tokens / 32))
        return kwargs_processed

    def post_process(self, res, max_new_tokens):

        response = json.loads(res.content.split(self.streaming_delimiter)[-2].decode())
        response['text'] = self.stem_message(response['text'], max_new_tokens)

        return response

    def _completion(self, prompt=None, **kwargs):

        prompt = self.get_prompt([{'role': 'user', 'content': prompt}])
        return self._completion_internal(prompt, **kwargs)

    def _chat_completion(self, messages=None, **kwargs):

        prompt = self.get_prompt(messages)
        return self._completion_internal(prompt, **kwargs)

    def _completion_internal(self, prompt, stop=None, stop_token_ids=None, cut_long_prompt=None,
                             echo=None, task=None, network=None, application=None, **kwargs):

        kwargs_processed = self.process_kwargs(prompt, **kwargs)
        d = dict(model=self.model, user=self.consumer, max_new_tokens=kwargs['max_new_tokens'],
                 prompt=prompt, temperature=kwargs['temperature'], stop_token_ids=stop_token_ids or self.stop_token_ids,
                 stop=stop or self.stop, cut_long_prompt=cut_long_prompt or self.cut_long_prompt,
                 echo=echo or self.echo, task=self.task or self.task, network=network or self.network,
                 application=application or self.application,
                 number_of_generations=kwargs_processed['number_of_generations'])

        res = requests.post(f"{self.protocol}://{self.normalized_hostname}/{self.worker_generate_stream_ep}",
                            headers=self.headers, json=d, verify=False, timeout=self.timeout)
        res = self.post_process(res, kwargs['max_new_tokens'])
        return CompletionObject(prompt=d['prompt'], kwargs=d, response=res)

    def verify_response(self, res):
        try:

            if res.response['error_code']:
                logger.warning(f"Model {self.model}: error_code={res.response['error_code']}")

            assert 'text' in res.response, f"Response does not contain 'text' key"

        except Exception as e:
            logger.error(f"Error in response: {res.response['error_code']}")
            raise e
        return True

    def extract_text(self, res):
        return res.response['text']

    def openai_format(self, res):
        return super().openai_format(res, tokens=res.response['tokens'],
                                     completion_tokens=res.response['completion_tokens'],
                                     total_tokens=res.response['total_tokens'])


class HuggingFaceLLM(BeamLLM):
    pipline_kwargs: Any
    input_device: Optional[str] = Field(None)
    eos_pattern: Optional[str] = Field(None)
    tokenizer_name: Optional[str] = Field(None)
    _text_generation_pipeline: Any = PrivateAttr(default=None)
    _conversational_pipeline: Any = PrivateAttr(default=None)

    def __init__(self, model, tokenizer=None, dtype=None, chat=False, input_device=None, compile=True, *args,
                 model_kwargs=None,
                 config_kwargs=None, pipline_kwargs=None, text_generation_kwargs=None, conversational_kwargs=None,
                 eos_pattern=None, **kwargs):

        kwargs['scheme'] = 'huggingface'
        kwargs['model'] = model
        super().__init__(*args, **kwargs)

        import transformers
        transformers.logging.set_verbosity_error()

        if model_kwargs is None:
            model_kwargs = {}

        if config_kwargs is None:
            config_kwargs = {}

        if pipline_kwargs is None:
            pipline_kwargs = {}

        if text_generation_kwargs is None:
            text_generation_kwargs = {}

        if conversational_kwargs is None:
            conversational_kwargs = {}

        self.pipline_kwargs = pipline_kwargs

        self.input_device = input_device
        self.eos_pattern = eos_pattern

        from transformers import AutoModelForCausalLM, AutoConfig

        self.config = AutoConfig.from_pretrained(model, trust_remote_code=True, **config_kwargs)
        self.tokenizer_name = tokenizer or model

        self.net = AutoModelForCausalLM.from_pretrained(model, trust_remote_code=True,
                                                        config=self.config, **model_kwargs)

        if compile:
            import torch
            self.net = torch.compile(self.net)

        self._text_generation_pipeline = transformers.pipeline('text-generation', model=self.net,
                                                               tokenizer=self.tokenizer, device=self.input_device,
                                                               return_full_text=False, **text_generation_kwargs)

        self._conversational_pipeline = transformers.pipeline('conversational', model=self.net,
                                                              tokenizer=self.tokenizer, device=self.input_device,
                                                              **conversational_kwargs)

    @lazy_property
    def tokenizer(self):
        from transformers import AutoTokenizer
        return AutoTokenizer.from_pretrained(self.tokenizer_name, trust_remote_code=True)

    def extract_text(self, res):

        res = res.response
        if type(res) is list:
            res = res[0]

        if type(res) is Conversation:
            res = res.generated_responses[-1]
        else:
            res = res['generated_text']

        if self.eos_pattern:
            res = res.split(self.eos_pattern)[0]

        return res

    @property
    def is_chat(self):
        return True

    @property
    def is_completions(self):
        return True

    def _completion(self, prompt=None, **kwargs):

        # pipeline = transformers.pipeline('text-generation', model=self.model,
        #                                  tokenizer=self.tokenizer, device=self.input_device, return_full_text=False)

        res = self._text_generation_pipeline(prompt, pad_token_id=self._text_generation_pipeline.tokenizer.eos_token_id,
                                             **self.pipline_kwargs)

        return CompletionObject(prompt=prompt, kwargs=kwargs, response=res)

    def _chat_completion(self, **kwargs):

        # pipeline = transformers.pipeline('conversational', model=self.model,
        #                                  tokenizer=self.tokenizer, device=self.input_device)

        res = self._conversational_pipeline(self.conversation,
                                             pad_token_id=self._conversational_pipeline.tokenizer.eos_token_id,
                                             **self.pipline_kwargs)

        return CompletionObject(prompt=self.conversation, kwargs=kwargs, response=res)


import warnings
import torch

from ..utils import check_type, as_tensor, as_numpy
from ..utils import check_type
from .universal_dataset import UniversalDataset


class LazyReplayBuffer(UniversalDataset):

    def __init__(self, size, device='cpu'):
        self.max_size = size
        self.size = 0
        self.ptr = 0
        self.target_device = device
        super().__init__(device=device)

    def build_buffer(self, x):
        return torch.zeros(self.size, *x.shape, device=self.target_device, dtype=x.dtype)

    def store(self, *args, **kwargs):

        if len(args) == 1:
            d = args[0]
        elif len(args):
            d = args
        else:
            d = kwargs

        if self.data is None:
            if isinstance(d, dict):
                self.data = {k: self.build_buffer(v) for k, v in d.items()}
                self.data_type = 'dict'
            elif isinstance(d, list) or isinstance(d, tuple):
                self.data = [self.build_buffer(v) for v in d]
                self.data_type = 'list'
            else:
                self.data = self.build_buffer(d)
                self.data_type = 'simple'

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if self.data_type == 'dict':
                for k, v in d.items():
                    self.data[k][self.ptr] = as_tensor(v, device=self.target_device)
            elif self.data_type == 'list':
                for i, v in enumerate(self.data):
                    self.data[i][self.ptr] = as_tensor(v, device=self.target_device)
            else:
                self.data[self.ptr] = as_tensor(d, device=self.target_device)

        self.ptr = (self.ptr + 1) % self.max_size
        self.size = min(self.size + 1, self.max_size)

    def reset(self):
        self.ptr = 0
        self.data = None
        self.size = 0

    def __len__(self):
        return self.size


class TransformedDataset(torch.utils.data.Dataset):
    def __init__(self, dataset, alg, *args, **kwargs):
        super().__init__()

        if type(dataset) != UniversalDataset:
            dataset = UniversalDataset(dataset)

        self.dataset = dataset
        self.alg = alg
        self.args = args
        self.kwargs = kwargs

    def __getitem__(self, ind):

        ind_type = check_type(ind, check_element=False)
        if ind_type.major == 'scalar':
            ind = [ind]

        ind, data = self.dataset[ind]
        dataset = UniversalDataset(data)
        res = self.alg.predict(dataset, *self.args, **self.kwargs)

        return ind, res.values



from concurrent.futures import ThreadPoolExecutor, Future
from functools import wraps

from ..core import MetaAsyncResult, MetaDispatcher
from ..utils import get_class_properties, lazy_property


class ThreadAsyncResult(MetaAsyncResult):
    def __init__(self, future: Future):
        super().__init__(future)
        self.future = future

    @property
    def value(self):
        return self.future.result()

    def wait(self, timeout=None):
        return self.future.result(timeout=timeout)

    def kill(self, no_restart=False):
        self.future.cancel()

    @property
    def is_ready(self):
        return self.future.done()

    @property
    def hex(self):
        return hex(id(self.future))

    def __repr__(self):
        return f"AsyncResult({self.str}, is_ready={self.is_ready}, is_success={self.is_success})"

    @property
    def state(self):
        if self.future.running():
            return "RUNNING"
        elif self.future.done():
            return "FINISHED"
        elif self.future.cancelled():
            return "CANCELLED"
        else:
            return "PENDING"


class ThreadedCluster:
    def __init__(self, max_workers=None):
        self.executor = ThreadPoolExecutor(max_workers=max_workers)

    def submit(self, fn, *args, **kwargs):
        future = self.executor.submit(fn, *args, **kwargs)
        return future

    def map(self, fn, *iterables):
        return self.executor.map(fn, *iterables)

    def shutdown(self, wait=True):
        self.executor.shutdown(wait=wait)


class ThreadedRemoteClass:
    def __init__(self, target_class, asynchronous=False, executor=None, properties=None):
        self._target_class = target_class
        self.properties = properties if properties is not None else []
        self.asynchronous = asynchronous
        if executor is None:
            self.executor = ThreadedCluster()
        else:
            self.executor = executor

    @property
    def target_class(self):
        if isinstance(self._target_class, ThreadAsyncResult):
            self._target_class = self._target_class.value
        return self._target_class

    def method_wrapper(self, method):
        def wrapper(*args, **kwargs):
            future = self.executor.submit(method, *args, **kwargs)
            if self.asynchronous:
                return ThreadAsyncResult(future)
            else:
                return future.result()
        return wrapper

    def __getattr__(self, item):

        if item in self.properties:
            future = self.executor.submit(self.target_class._get_property, item)
            if self.asynchronous:
                return ThreadAsyncResult(future)
            else:
                return future.result()

        attr = getattr(self.target_class, item)
        if callable(attr):
            return self.method_wrapper(attr)
        return attr

    def __call__(self, *args, **kwargs):
        return self.method_wrapper(self.target_class.__call__)(*args, **kwargs)


class ThreadedDispatcher(MetaDispatcher, ThreadedCluster):
    def __init__(self, obj, *routes, max_workers=None, asynchronous=True, **kwargs):
        MetaDispatcher.__init__(self, obj, *routes, asynchronous=asynchronous, **kwargs)
        ThreadedCluster.__init__(self, max_workers=max_workers)

        if self.type == 'function':
            self.call_function = self.threaded_function_wrapper(self.obj)
        elif self.type == 'instance':
            if hasattr(self.obj, '__call__'):
                self.call_function = self.threaded_function_wrapper(self.obj.__call__)
            for route in self.routes:
                if hasattr(self.obj, route):
                    method = getattr(self.obj, route)
                    self._routes_methods[route] = self.threaded_function_wrapper(method)
        elif self.type == 'class':
            self.call_function = self.threaded_function_wrapper(self.factory_class_wrapper(self.obj))
        else:
            raise ValueError(f"Unknown type: {self.type}")

    @lazy_property
    def route_methods(self):
        return self._routes_methods

    @staticmethod
    def factory_class_wrapper(cls):

        class RemoteClassWrapper(cls):
            def _get_property(self, prop):
                return getattr(self, prop)

        def wrapper(*args, **kwargs):
            res = RemoteClassWrapper(*args, **kwargs)
            return res

        return wrapper

    def threaded_function_wrapper(self, func, asynchronous=None, bypass=False):

        if asynchronous is None:
            asynchronous = self.asynchronous
        @wraps(func)
        def wrapper(*args, **kwargs):
            future = self.executor.submit(func, *args, **kwargs)
            if bypass:
                return future
            elif asynchronous:
                return ThreadAsyncResult(future)
            else:
                return future.result()
        return wrapper

    def __call__(self, *args, **kwargs):
        res = self.call_function(*args, **kwargs)
        if self.type == 'class':
            properties = get_class_properties(self.obj)
            return ThreadedRemoteClass(res, asynchronous=self.asynchronous, properties=properties)
        else:
            return res